# -*- coding: utf-8 -*-

# Copyright (c) 2006-2010, Rectorate of the University of Freiburg
# Copyright (c) 2010, Andreas W. Liehr (liehr@users.sourceforge.net)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

u"""
"""
from pyphant.mplbackend import ensure_mpl_backend
ensure_mpl_backend('wxagg')

import os, os.path, pkg_resources
from pyphant.core.Helpers import getPyphantPath
LOGDIR = getPyphantPath()
import logging
from logging.handlers import MemoryHandler
logging.basicConfig(level=logging.NOTSET,
                    filename=os.path.join(LOGDIR, u'pyphant.log'),
                    filemode='w',
                    format="%(asctime)s - %(levelname)s:%(name)s:%(thread)"\
                    "d:%(module)s.%(funcName)s(l %(lineno)d):%(message)s")
console = logging.StreamHandler()
console.setLevel(logging.WARNING)
pdmh = MemoryHandler(1000, flushLevel=logging.CRITICAL + 1)
pdmh.setLevel(logging.WARNING)
logging.getLogger('').addHandler(pdmh)
logging.getLogger('').addHandler(console)

import sys
import wx
import wx.aui
import sogl
import pyphant.wxgui2.paramvisualization.ParamVisReg as ParamVisReg
from pyphant.core.H5FileHandler import H5FileHandler
import WorkerRepository
import ConfigureFrame
import platform
from pyphant.core.KnowledgeManager import KnowledgeManager
import webbrowser
pltform = platform.system()


class wxPyphantApplication(wx.App):
    def __init__(self, pathToRecipe=None):
        self.pathToRecipe = pathToRecipe
        wx.App.__init__(self)

    def OnInit(self):
        self._logger = logging.getLogger("pyphant")
        sys.excepthook = self.excepthook
        sogl.SOGLInitialize()
        self._knowledgeNode = None
        self._paramVisReg = ParamVisReg.ParamVisReg()
        self.splash = mySplashScreen(self)
        self.splash.Center()
        self.splash.Show()
        return True

    def excepthook(self, type, value, trace):
        self._logger.error(u"An unhandled exception occured.",
                           exc_info=(type, value, trace))
        sys.__excepthook__(type, value, trace)

    def getMainFrame(self):
        return self._frame

    def configureWorker(self, worker):
        configureFrame = ConfigureFrame.ConfigureFrame(self._frame,
                                                       self._paramVisReg,
                                                       worker)
        if configureFrame.ShowModal() == wx.ID_OK:
            configureFrame.applyAll()

    def editCompositeWorker(self, worker):
        self._frame.editCompositeWorker(worker)


class AbortRecipeCreation(Exception):
    pass


class wxPyphantFrame(wx.Frame):

    ID_WINDOW_TOP = 100
    ID_WINDOW_LEFT = 101
    ID_WINDOW_RIGHT = 102
    ID_WINDOW_BOTTOM = 103
    ID_CLOSE_COMPOSITE_WORKER = wx.NewId()
    ID_UPDATE_PYPHANT = wx.NewId()
    ID_VIEW_WORKERREP = wx.NewId()
    ID_VIEW_LOGFILE = wx.NewId()
    ID_KM_URL = wx.NewId()
    ID_KM_LOCAL = wx.NewId()
    ID_KM_SHARE = wx.NewId()

    def __init__(self, _wxPyphantApp):
        version = pkg_resources.get_distribution("pyphant").version
        self.titleStr = "wxPyphant %s | Recipe: %s" % (version, "%s")
        wx.Frame.__init__(self, None, -1, self.titleStr % "None",
                          size=(640,480))
        self._statusBar = self.CreateStatusBar()
        self._wxPyphantApp = _wxPyphantApp
        self._initMenuBar()
        self._initWorkerRep()
        self.recipeState = None
        try:
            self.onOpenCompositeWorker(None)
        except AbortRecipeCreation:
            self.shutdown()
            return
        except IOError as error:
            self._wxPyphantApp._logger.error(
                error.message,
                exc_info=sys.exc_info()
                )
            self.shutdown()
            return
        self._initAui()
        self.compositeWorkerStack = []

    def _initWorkerRep(self):
        try:
            # adds Emd5Src worker to WorkerRepository
            from pyphant.core.Emd5Src import Emd5Src
            Emd5Src()
            self._workerRepository = WorkerRepository.WorkerRepository(
                self, self.ID_WINDOW_RIGHT, wx.DefaultPosition,
                wx.Size(220, -1))
            self._workerRepository.Expand(self._workerRepository.RootItem)
        except:
            import sys
            self._wxPyphantApp._logger.error(
                u"An exception occured while loading the toolboxes.",
                exc_info=sys.exc_info())
            wx.MessageBox(
                "An error has occurred while importing "\
                "the toolboxes.\nPlease investigate the logfile "\
                "for further details.\nThe logfile is located at %s\n"\
                "You may also try to update and restart wxPyphant."\
                % os.path.join(LOGDIR, 'pyphant.log'),
                "Toolbox Error!")
            self._workerRepository = wx.TreeCtrl(self)

    def _initAui(self):
        self._auiManager = wx.aui.AuiManager(self)
        self._logpane = wx.TextCtrl(
            self, -1, "", wx.DefaultPosition, wx.Size(640, 200),
            wx.NO_BORDER | wx.TE_MULTILINE | wx.TE_READONLY
            )
        class TextCtrlStream(object):
            def __init__(self, textctrl):
                self.textctrl = textctrl

            def write(self, msg):
                try:
                    wx.CallAfter(self.textctrl.WriteText, msg)
                except wx.PyDeadObjectError:
                    pass
        handler = logging.StreamHandler(TextCtrlStream(self._logpane))
        handler.setFormatter(logging.Formatter(
            "%(asctime)s - %(levelname)s:%(name)s:%(thread)"\
            "d:%(module)s.%(funcName)s(l %(lineno)d):%(message)s"))
        handler.setLevel(logging.WARNING)
        global pdmh
        pdmh.setTarget(handler)
        pdmh.close()
        logging.getLogger('').removeHandler(pdmh)
        del pdmh
        logging.getLogger('').addHandler(handler)
        self._auiManager.AddPane(self._logpane, wx.BOTTOM, 'Logfile @ %s'\
                                 % os.path.join(LOGDIR, 'pyphant.log'))
        self._auiManager.AddPane(self._workerRepository, wx.RIGHT,
                                 'Worker Repository')
        self._auiManager.AddPane(self._remainingSpace, wx.CENTER, 'Main')
        self._auiManager.GetPane(self._workerRepository)
        self._auiManager.Update()

    def onOpenCompositeWorker(self, event):
        if not self._wxPyphantApp.pathToRecipe:
            if pltform == 'Linux' or pltform == 'Darwin':
                osMessage = "Choose an existing recipe or cancel to create "\
                            "a new recipe"
            elif pltform == 'Windows':
                osMessage = "Choose existing recipe to open or name a new "\
                            "recipe to create"
            else:
                raise OSError, "Operating System %s not supported!" % pltform
            wc = "Pyphant Recipe(*.h5)|*.h5"
            dlg = wx.FileDialog(
                self, message=osMessage, defaultDir=os.getcwd(),
                defaultFile="", wildcard=wc, style=wx.OPEN
                )
            if dlg.ShowModal() == wx.ID_OK:
                self._wxPyphantApp.pathToRecipe = dlg.GetPath()
            else:
                dlg.Destroy()
                dlg = wx.FileDialog(self, message='Create a new recipe',
                                    defaultDir=os.getcwd(), defaultFile="",
                                    wildcard=wc, style=wx.SAVE)
                if dlg.ShowModal() == wx.ID_OK:
                    path = dlg.GetPath()
                    if not path[:-3] == '.h5':
                        path += '.h5'
                    self._wxPyphantApp.pathToRecipe = path
                else:
                    dlg.Destroy()
                    raise AbortRecipeCreation
            dlg.Destroy()
        import PyphantCanvas
        if self._wxPyphantApp.pathToRecipe[-3:] == '.h5':
            if os.path.exists(self._wxPyphantApp.pathToRecipe):
                try:
                    with H5FileHandler(self._wxPyphantApp.pathToRecipe, 'r') \
                             as handler:
                        recipe = handler.loadRecipe()
                    self._remainingSpace = PyphantCanvas.PyphantCanvas(
                        self, recipe)
                except:
                    self._wxPyphantApp._logger.error(
                        u"An exception occured while loading a recipe.",
                        exc_info=sys.exc_info())
                    wx.MessageBox(
                        "An error has occurred while opening "\
                        "the recipe.\nRecipe has been set to an "\
                        "empty file in order to prevent data loss.\n"\
                        "Please investigate the logfile "\
                        "for further details.\nThe logfile is located at %s"\
                        % os.path.join(LOGDIR, 'pyphant.log'),
                        "Recipe broken, unknown format or outdated!")
                    self._wxPyphantApp.pathToRecipe += ".error.h5"
                    self._remainingSpace = PyphantCanvas.PyphantCanvas(self)
            else:
                self._remainingSpace = PyphantCanvas.PyphantCanvas(self)
            from pyphant.core.WebInterface import shorten
            self.SetTitle(self.titleStr \
                          % shorten(self._wxPyphantApp.pathToRecipe, 30, 30))
        else:
            raise IOError('Unknown file format in file "%s"'\
                              % self._wxPyphantApp.pathToRecipe)
        self.recipeState = 'clean'
        self._remainingSpace.diagram.recipe.registerListener(
            self.recipeChanged
            )

    def recipeChanged(self, event):
        self.recipeState = 'dirty'

    def onSaveCompositeWorker(self, event=None):
        with H5FileHandler(self._wxPyphantApp.pathToRecipe, 'w') as handler:
            handler.saveRecipe(self._remainingSpace.diagram.recipe,
                               self._fileMenu.IsChecked(wx.ID_FILE4))
        self.recipeState = 'clean'

    def onSaveAsCompositeWorker(self, event=None):
        msg = "Select file to save recipe."
        wc = "Pyphant recipe (*.h5)|*.h5"
        dlg = wx.FileDialog(self, message=msg, defaultDir=os.getcwd(),
                            defaultFile="", wildcard=wc, style=wx.SAVE)
        if dlg.ShowModal() == wx.ID_OK:
            filename = dlg.GetPath()
            if not filename.endswith(".h5"):
                filename += ".h5"
            with H5FileHandler(filename, 'w') as handler:
                handler.saveRecipe(
                    self._remainingSpace.diagram.recipe,
                    self._fileMenu.IsChecked(wx.ID_FILE4))
            self._wxPyphantApp.pathToRecipe = filename
            self.recipeState = 'clean'
            from pyphant.core.WebInterface import shorten
            self.SetTitle(self.titleStr % shorten(filename, 30, 30))
        else:
            dlg.Destroy()

    def _initMenuBar(self):
        self._menuBar = wx.MenuBar()
        self._fileMenu = wx.Menu()
        self._fileMenu.AppendCheckItem(wx.ID_FILE4, "Save &results\tCTRL+r")
        self._fileMenu.Check(wx.ID_FILE4, True)
        self._fileMenu.Append(wx.ID_SAVE, "&Save\tCTRL+s")
        self._fileMenu.Append(wx.ID_SAVEAS, "Save &as\tCTRL+a")
        self._fileMenu.Append(wx.ID_EXIT, "E&xit")
        self._menuBar.Append(self._fileMenu, "&File")
        self._closeCompositeWorker = wx.Menu()
        self._closeCompositeWorker.Append(self.ID_CLOSE_COMPOSITE_WORKER,
                                          "&Close Composite Worker")
        self._menuBar.Append(self._closeCompositeWorker,
                              "&Close Composite Worker")
        self._updateMenu = self.createUpdateMenu()
        self._menuBar.Append(self._updateMenu, "&Update")
        self._viewMenu = self.createViewMenu()
        self._menuBar.Append(self._viewMenu, "&View")
        self._kmanagerMenu = self.createKmanagerMenu()
        self._menuBar.Append(self._kmanagerMenu, "&Knowledge Manager")
        self.SetMenuBar(self._menuBar)
        self._menuBar.EnableTop(1, False)
        self.Bind(wx.EVT_MENU, self.onSaveCompositeWorker, id=wx.ID_SAVE)
        self.Bind(wx.EVT_MENU, self.onSaveAsCompositeWorker, id=wx.ID_SAVEAS)
        self.Bind(wx.EVT_CLOSE, self.onClose)
        self.Bind(wx.EVT_MENU, self.onQuit, id=wx.ID_EXIT)
        self.Bind(wx.EVT_MENU, self.onCloseCompositeWorker,
                  id=self.ID_CLOSE_COMPOSITE_WORKER)

    def createUpdateMenu(self):
        updateMenu = wx.Menu()
        updateMenu.Append(self.ID_UPDATE_PYPHANT, "Update &Pyphant")
        self.Bind(wx.EVT_MENU, self.onUpdatePyphant, id=self.ID_UPDATE_PYPHANT)
        self.updateIds = { self.ID_UPDATE_PYPHANT : 'pyphant' }
        for toolbox in pkg_resources.iter_entry_points("pyphant.workers"):
            dist = toolbox.dist
            nId = wx.NewId()
            self.updateIds[nId] = dist.key
            updateMenu.Append(nId, "Update %s (%s)"\
                              % (dist.project_name, dist.version))
            self.Bind(wx.EVT_MENU, self.onUpdatePyphant, id=nId)
        return updateMenu

    def createViewMenu(self):
        viewMenu = wx.Menu()
        viewMenu.Append(self.ID_VIEW_WORKERREP, "&Worker Repository")
        viewMenu.Append(self.ID_VIEW_LOGFILE, "&Logfile")
        self.Bind(wx.EVT_MENU, self.onWorkerRep, id=self.ID_VIEW_WORKERREP)
        self.Bind(wx.EVT_MENU, self.onLogfile, id=self.ID_VIEW_LOGFILE)
        return viewMenu

    def createKmanagerMenu(self):
        kmanagerMenu = wx.Menu()
        kmanagerMenu.Append(self.ID_KM_URL, "Import HDF5 or FMF from &URL")
        kmanagerMenu.Append(self.ID_KM_LOCAL, "Import &local HDF5 or FMF file")
        kmanagerMenu.Append(self.ID_KM_SHARE, "Start/pause sharing &knowledge")
        self.Bind(wx.EVT_MENU, self.onImportURL, id=self.ID_KM_URL)
        self.Bind(wx.EVT_MENU, self.onImportLocal, id=self.ID_KM_LOCAL)
        self.Bind(wx.EVT_MENU, self.onShare, id=self.ID_KM_SHARE)
        return kmanagerMenu

    def onUpdatePyphant(self, event):
        packageName = self._updateMenu.FindItemById(event.Id)\
                      .GetItemLabelText()[7:]
        cpt = u"Update Manager"
        msg = u"Trying to update package '%s'.\nYou will be notified "\
              "when the process has finished.\n"\
              "Please press 'OK' now to begin." % packageName
        dlg = wx.MessageDialog(self, msg, cpt, style=wx.OK|wx.CANCEL)
        dlgid = dlg.ShowModal()
        dlg.Destroy()
        if dlgid != wx.ID_OK:
            return
        try:
            import pyphant.core.UpdateManager
            error = pyphant.core.UpdateManager.updatePackage(
                self.updateIds[event.Id])
        except Exception, exc:
            error = "%s:\n%s" % (exc.__class__.__name__, exc.message)
        if error is not None and len(error) > 0:
            msg = u"An error occured during the update of '%s':\n%s"\
                  % (packageName, error)
        else:
            msg = u"Finished updating '%s'." % packageName
        dlg = wx.MessageDialog(self, msg, cpt, style=wx.OK)
        dlg.ShowModal()
        dlg.Destroy()

    def onWorkerRep(self, event):
        wrpane = self._auiManager.GetPane(self._workerRepository)
        wrpane.Show(True)
        self._auiManager.Update()

    def onLogfile(self, event):
        logpane = self._auiManager.GetPane(self._logpane)
        logpane.Show(True)
        self._auiManager.Update()

    def onQuit(self,event):
        self.Close()

    def onClose(self, event):
        dlgid = None
        if self.recipeState != 'clean':
            cpt = "Save changed recipe?"
            msg = "The recipe has changed since the last saving.\n"\
                  "Do you want to save before terminating?"
            dlg = wx.MessageDialog(
                self, msg, cpt, style=wx.YES|wx.NO|wx.CANCEL|wx.ICON_QUESTION)
            dlgid = dlg.ShowModal()
            if dlgid == wx.ID_YES:
                self.onSaveCompositeWorker()
            dlg.Destroy()
        if dlgid != wx.ID_CANCEL:
            self.shutdown()

    def shutdown(self):
        try:
            self._wxPyphantApp._knowledgeNode.stop()
        except AttributeError:
            pass
        try:
            self._auiManager.UnInit()
        except AttributeError:
            pass
        self.Destroy()

    def editCompositeWorker(self, worker):
        import PyphantCanvas
        self.compositeWorkerStack.append(self._remainingSpace)
        self._remainingSpace = PyphantCanvas.PyphantCanvas(self, worker)
        self._remainingSpace.diagram.recipe.registerListener(
            self.recipeChanged
            )
        self._menuBar.EnableTop(1, True)

    def onCloseCompositeWorker(self, event):
        self._remainingSpace.Destroy()
        self._remainingSpace = self.compositeWorkerStack.pop()
        if len(self.compositeWorkerStack) == 0:
            self._menuBar.EnableTop(1, False)

    def onImportURL(self, event):
        cpt = "Import HDF5 or FMF from URL"
        msg = "Enter a URL to a valid HDF5 or FMF file "\
              "(e.g. http://www.example.org/data.h5).\n"\
              "The file is stored permanently in your home directory in the "\
              ".pyphant directory\nand all DataContainers contained in that "\
              "file are available by using the\nSCSource or DCSource "\
              "Worker even after "\
              "restarting wxPyphant.\nHTTP redirects are resolved "\
              "automatically, i.e. DOIs are supported as well."
        dlg = wx.TextEntryDialog(self, msg, cpt)
        dlgid = dlg.ShowModal()
        if dlgid != wx.ID_CANCEL:
            url = dlg.GetValue()
            cpt2 = "Info"
            msg2 = "Successfully imported DataContainers from\n'%s'"\
                   % (url ,)
            km = KnowledgeManager.getInstance()
            try:
                km.registerURL(url)
            except Exception:
                cpt2 = "Error"
                msg2 = "'%s' is not a valid URL to a HDF5 or FMF file."\
                       % (url, )
            finally:
                dlg2 = wx.MessageDialog(self, msg2, cpt2, wx.OK)
                dlg2.ShowModal()
                dlg2.Destroy()
        dlg.Destroy()

    def onImportLocal(self, event):
        msg = "Select HDF5 or FMF file to import DataContainer(s) from."
        wc = "*.h5, *.hdf, *.hdf5, *.fmf|*.h5;*.hdf;*.hdf5;*.fmf"
        dlg = wx.FileDialog(self, message=msg, defaultDir=os.getcwd(),
                            defaultFile="", wildcard=wc, style=wx.OPEN)
        if dlg.ShowModal() == wx.ID_OK:
            filename = dlg.GetPath()
            url = 'file://' + os.path.realpath(filename)
            km = KnowledgeManager.getInstance()
            cpt2 = "Info"
            msg2 = "Successfully imported DataContainer(s) from\n'%s'"\
                   % (filename ,)
            try:
                km.registerURL(url)
            except Exception:
                cpt2 = "Error"
                msg2 = "'%s' is not a valid HDF5 or FMF file.\n"\
                       "(Tried to import from '%s')" % (filename, url)
            finally:
                dlg2 = wx.MessageDialog(self, msg2, cpt2, wx.OK)
                dlg2.ShowModal()
                dlg2.Destroy()
        else:
            dlg.Destroy()

    def onShare(self, event):
        cpt = "Share Knowledge"
        msg = ""
        if self._wxPyphantApp._knowledgeNode is None:
            try:
                logg = self._wxPyphantApp._logger
                from pyphant.core.KnowledgeNode import get_kn_autoport
                ports = [8080] + range(48621, 48771)
                self._wxPyphantApp._knowledgeNode = get_kn_autoport(
                    ports, logg, start=True, web_interface=True)
                url = self._wxPyphantApp._knowledgeNode.url
                msg += "Knowledge node is listening @ %s.\n"\
                       "Sharing is experimental and therefore restric"\
                       "ted\nto the loopback interface." % url
                webbrowser.open_new(url)
            except Exception, exep:
                msg += "Could not start web server."
                from socket import error as socket_error
                if isinstance(exep, socket_error):
                    try:
                        #Python 2.6
                        eno = exep.errno
                    except AttributeError:
                        #Python 2.5
                        eno = exep.args[0]
                    from errno import EADDRINUSE
                    if eno == EADDRINUSE:
                        msg += "\nReason: Could not find a free port."\
                               "\n(You may stop other applications or "\
                               "wait for the OS\nto free some ports.)"
        elif not self._wxPyphantApp._knowledgeNode.app.serve:
            msg += "Resumed sharing."
            self._wxPyphantApp._knowledgeNode.app.serve = True
        else:
            self._wxPyphantApp._knowledgeNode.app.serve = False
            msg += "Disabled sharing."
        dlg = wx.MessageDialog(self, msg, cpt, wx.OK)
        dlg.ShowModal()
        dlg.Destroy()


class mySplashScreen(wx.Frame):
    def __init__(self, parent):
        self.parent = parent
        wx.Frame.__init__(self, None, -1, "Shaped Window",
                         style =
                           wx.FRAME_SHAPED
                         | wx.SIMPLE_BORDER
                         | wx.FRAME_NO_TASKBAR
                         | wx.STAY_ON_TOP
                         )

        self.hasShape = False
        self.delta = (0,0)

        self.Bind(wx.EVT_RIGHT_UP,      self.OnExit)
        self.Bind(wx.EVT_PAINT,         self.OnPaint)

        self.timer = wx.Timer(self)
        self.timer.Start(5000,oneShot=True)
        self.Bind(wx.EVT_TIMER, self.OnExit)

        import StringIO
        import base64
        import pyphantLogo

        png = base64.decodestring(pyphantLogo.pic_b64)
        stream = StringIO.StringIO(png)
        wxImage = wx.ImageFromStream(stream)
        wxImage.ConvertAlphaToMask(10)
        self.bmp = wxImage.ConvertToBitmap()

        w, h = self.bmp.GetWidth(), self.bmp.GetHeight()
        self.SetClientSize( (w, h) )

        if wx.Platform != "__WXMAC__":
            # wxMac clips the tooltip to the window shape, YUCK!!!
            self.SetToolTipString("Loading Pyphant")

        if wx.Platform == "__WXGTK__":
            # wxGTK requires that the window be created before you can
            # set its shape, so delay the call to SetWindowShape until
            # this event.
            self.Bind(wx.EVT_WINDOW_CREATE, self.SetWindowShape)
        else:
            # On wxMSW and wxMac the window has already been created,
            # so go for it.
            self.SetWindowShape()

        dc = wx.ClientDC(self)
        dc.DrawBitmap(self.bmp, 0,0, True)

    def SetWindowShape(self, *evt):
        # Use the bitmap's mask to determine the region
        r = wx.RegionFromBitmap(self.bmp)
        self.hasShape = self.SetShape(r)

    def OnPaint(self, evt):
        dc = wx.PaintDC(self)
        dc.DrawBitmap(self.bmp, 0,0, True)

    def OnExit(self, evt):
        self.Hide()
        # wxPyphantFrame is the main frame.
        self.parent._frame = wxPyphantFrame(self.parent)
        self.parent._frame.Show()

        # The program will freeze without this line.
        evt.Skip()  # Make sure the default handler runs too...
        self.Destroy()


import optparse

def startWxPyphant():
    usage = "usage: %prog [options] pathToRecipe"
    version = pkg_resources.get_distribution("pyphant").version
    parser = optparse.OptionParser(usage, version=version)
    (options, args) = parser.parse_args()

    if len(args) > 0:
        pathToRecipe = args[0]
    else:
        pathToRecipe = None
    wxPyphantApp = wxPyphantApplication(pathToRecipe)
    wxPyphantApp.MainLoop()



# -*- coding: utf-8 -*-

# Copyright (c) 2006-2007, Rectorate of the University of Freiburg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

u"""
"""


import wx
import sogl
import PyphantDiagram
import pyphant.core.CompositeWorker as CompositeWorker



class PyphantCanvas(sogl.ShapeCanvas):
    def __init__(self,parent,recipe=CompositeWorker.CompositeWorker(), id=-1,pos=wx.DefaultPosition,
                 size=wx.DefaultSize,style=wx.NO_BORDER,name="PyphantCanvas"):
        sogl.ShapeCanvas.__init__(self,parent, id, pos, size, style, name)
        #self.SetVirtualSize((1100,1000))
        self.SetScrollRate(20,20)
        self.dropTarget = CanvasDropTarget(self)
        self.SetDropTarget(self.dropTarget)
        self.diagram = PyphantDiagram.PyphantDiagram(self, recipe)

    def getDC(self):
        return wx.ClientDC(self)

#    def OnBeginDragLeft(self, x, y, keys=0):
#        print "A"
#        print self._draggedShape


class CanvasDropTarget(wx.PyDropTarget):
    def __init__(self, parent):
        wx.PyDropTarget.__init__(self)
        self.parent = parent
        self.df = wx.CustomDataFormat('PYPHANT_WORKER')
        self.data = wx.CustomDataObject(self.df)
        self.SetDataObject(self.data)

    def OnDrop(self, x, y):
        return True

    def OnData(self, x, y, d):
        if self.GetData():
            dump = self.data.GetData()
            self.parent.diagram.addWorker(dump, (x, y))
        #self.parent.Refresh()
        return d



# -*- coding: utf-8 -*-

# Copyright (c) 2006-2013, Rectorate of the University of Freiburg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


def ensure_mpl_backend(backend=None):
    """
    Tries to select the matplotlib backend.

    Raises EnvironmentError, if the desired backend is valid but could not be selected.
    If `backend` is None, either 'agg' or 'wxagg' is chosen, depending on whether
    a display is present (on Linux and Darwin).
    For other platforms, 'wxagg' is always the default.
    Returns the selected backend as given by matplotlib.get_backend().
    """
    if backend is None:
        import platform
        import os
        if platform.system() in ('Linux', 'Darwin') and not 'DISPLAY' in os.environ:
            backend = 'agg'
        else:
            backend = 'wxagg'
    import matplotlib
    active_backend = matplotlib.get_backend()
    if active_backend.lower() != backend.lower():
        matplotlib.use(backend)
        active_backend = matplotlib.get_backend()
        if active_backend.lower() != backend.lower():
            raise EnvironmentError(
                "Could not select matplotlib backend '%s' ('%s' is active)!" \
                % (backend, active_backend)
                )
    return active_backend



# -*- coding: utf-8 -*-

# Copyright (c) 2008-2009, Rectorate of the University of Freiburg
# Copyright (c) 2009-2011, Andreas W. Liehr (liehr@users.sourceforge.net)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

u"""
"""


import mx.DateTime.ISO
from pyphant.quantities import Quantity
from pyphant.quantities.PhysicalQuantities import PhysicalQuantity

import logging
_logger = logging.getLogger("pyphant")

def str2unit(unitStr, FMFversion='1.1'):
    """return float or Quantity instance

    The function str2unit is a factory, which either returns a float
    or a Quantity instance from a given string.
    Because the definition of units and physical constants is not unique for
    FMFversion 1.0 (http://arxiv.org/abs/0904.1299) and
    FMFversion 1.1 (http://dx.doi.org/10.1016/j.cpc.2009.11.014),
    the result of str2unit() depends on FMFversion.
    """
    if FMFversion not in ['1.0', '1.1']:
        raise ValueError('FMFversion %s not supported.' % (FMFversion, ))
    # Deal with exceptional units like '%' or 'a.u.'
    if unitStr.endswith('%'):
        if len(unitStr.strip()) == 1:
            return 0.01
        else:
            return float(unitStr[:-1]) / 100.0
    elif unitStr.endswith('a.u.'):
        if len(unitStr.strip()) == 4:
            return 1.0
        else:
            return float(unitStr[:-4])
    # Prepare conversion to quantity
    if unitStr.startswith('.'):
        unitStr = '0' + unitStr
    elif not (unitStr[0].isdigit() or unitStr[0] == '-'):
        unitStr = '1' + unitStr
    # Convert input to quantity or float
    unitStr = unitStr.replace('^', '**')
    try: #FMFversion == '1.1'
        unit = Quantity(unitStr.encode('utf-8'))
    except:
        unit = None
    if FMFversion == '1.0':
        try:
            unit1_0 = PhysicalQuantity(unitStr.encode('utf-8'))
            unit1_1 = Quantity(str(unit1_0.inBaseUnits()))
        except:
            unit1_1 = None
        if isinstance(unit1_1, Quantity): # Unit exists in 1.0
            if isinstance(unit, Quantity): # Unit also exists in 1.1
                if unit.isCompatible(unit1_1.unit):
                    # Interpretation of unit has not changed
                    unit = unit1_1.inUnitsOf(unit.unit)
                else:
                    unit = unit1_1
                    _logger.warn('Usage of old unit "%s" required '
                                 'conversion to base units.' % (unitStr, ))
            else:
                unit = unit1_1
                _logger.warn('Usage of old unit "%s" required '
                             'conversion to base units.' % (unitStr, ))
    if unit is None:
        try:
            if 'j' in unitStr:
                unit = complex(unitStr)
            else:
                unit = float(unitStr)
        except:
            raise ValueError("Unit %s cannot be interpreted." % (unitStr, ))
    return unit

def parseQuantity(value, FMFversion='1.1'):
    import re
    pm = re.compile(ur"(?:\\pm|\+-|\+/-)")
    try:
        value, error = [s.strip() for s in pm.split(value)]
    except:
        error = None
    if value.startswith('('):
        value = float(value[1:])
        error, unit = [s.strip() for s in error.split(')')]
        unit = str2unit(unit,FMFversion)
        value *= unit
    else:
        value = str2unit(value,FMFversion)
    if error != None:
        if error.endswith('%'):
            error = value * float(error[:-1]) / 100.0
        else:
            try:
                error = float(error) * unit
            except:
                error = str2unit(error, FMFversion)
    return value, error

def parseVariable(oldVal, FMFversion='1.1'):
    shortname, value = tuple([s.strip() for s in oldVal.split('=')])
    value, error = parseQuantity(value, FMFversion)
    return (shortname, value, error)

def parseDateTime(value, FMFversion='1.1'):
    """
    >>>parseDateTime('2004-08-21 12:00:00+-12hr')
    (Quantity(731814.5, 'd'), Quantity(0.5, 'd'))
    >>>parseDateTime('2004-08-21 12:00:00')
    (Quantity(731814.5, 'd'), None)
    """
    datetimeWithError = value.split('+-')
    if len(datetimeWithError) == 2:
        datetime = mx.DateTime.ISO.ParseAny(datetimeWithError[0])
        uncertainty = parseQuantity(datetimeWithError[1], FMFversion)[0]
        if uncertainty.isCompatible('h'):
            _logger.warning(
                "The uncertainty of timestamp %s has the unit 'h', "
                "which is deprecated. "
                "The correct abbreviation for hour is 'hr'." % (value, ))
            uncertainty = uncertainty * Quantity('1hr/h')
        error = uncertainty.inUnitsOf('d')
    else:
        datetime = mx.DateTime.ISO.ParseAny(value)
        error = None
    days, seconds = datetime.absvalues()
    return (Quantity(days, 'd') + Quantity(seconds, 's'), error)



# -*- coding: utf-8 -*-

# Copyright (c) 2009, Rectorate of the University of Freiburg
# Copyright (c) 2009-2010, Andreas W. Liehr (liehr@users.sourceforge.net)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



import unittest, numpy
from pyphant.quantities import Quantity
"""
    >>>Quantity('1V')
    Quantity(1.0,'d'), Quantity(0.5,'d'))
    >>>parseDateTime('2004-08-21 12:00:00')
    (Quantity(731814.5,'d'), None)
"""
class TestQuantity(unittest.TestCase):
    def setUp(self):
        self.quants = []
        self.quants.append(Quantity("1000 m"))
        self.quants.append(Quantity("1 km"))
        self.quants.append(Quantity("1000.1 m"))
        self.quants.append(0)
        self.quants.append(None)
        self.quants.append(Quantity("1000 s"))

    def testTextualQuantitySpecification(self):
        self.assertEqual(Quantity('1V'),
                         Quantity(1.0,'V')
                         )

    def testUnicodeQuantitySpecification(self):
        self.assertEqual(Quantity(u'1V'),
                         Quantity(1.0,'V')
                         )
        self.assertEqual(Quantity('1V'.encode('utf-8')),
                         Quantity(1.0,'V')
                         )

    def compare(self, quant0, cop, quant1, result):
        if cop == '==':
            expression = quant0 == quant1
        elif cop == '!=':
            expression = quant0 != quant1
        elif cop == '<=':
            expression = quant0 <= quant1
        elif cop == '>=':
            expression = quant0 >= quant1
        elif cop == '<':
            expression = quant0 < quant1
        elif cop == '>':
            expression = quant0 > quant1
        if expression != result:
            print "%s %s %s = %s, should be %s" % (quant0, cop, quant1,
                                                   expression, result)
        self.assertEqual(expression, result)

    def matrixTest(self, matrix, cop):
        for index0, quant0 in enumerate(self.quants):
            for index1, quant1 in enumerate(self.quants):
                self.compare(quant0, cop, quant1, matrix[index0][index1])

    def testComparisonOperationEqual(self):
        matrix = [[1, 1, 0, 0, 0, 0],
                  [1, 1, 0, 0, 0, 0],
                  [0, 0, 1, 0, 0, 0],
                  [0, 0, 0, 1, 0, 0],
                  [0, 0, 0, 0, 1, 0],
                  [0, 0, 0, 0, 0, 1]]
        self.matrixTest(matrix, '==')

    def testComparisonOperationNotEqual(self):
        matrix = [[0, 0, 1, 1, 1, 1],
                  [0, 0, 1, 1, 1, 1],
                  [1, 1, 0, 1, 1, 1],
                  [1, 1, 1, 0, 1, 1],
                  [1, 1, 1, 1, 0, 1],
                  [1, 1, 1, 1, 1, 0]]
        self.matrixTest(matrix, '!=')

    def testComparisonOperationLessOrEqual(self):
        matrix = [[1, 1, 1, 0, 0, 0],
                  [1, 1, 1, 0, 0, 0],
                  [0, 0, 1, 0, 0, 0],
                  [0, 0, 0, 1, 0, 0],
                  [1, 1, 1, 1, 1, 1],
                  [0, 0, 0, 0, 0, 1]]
        self.matrixTest(matrix, '<=')

    def testComparisonOperationGreaterOrEqual(self):
        matrix = [[1, 1, 0, 0, 1, 0],
                  [1, 1, 0, 0, 1, 0],
                  [1, 1, 1, 0, 1, 0],
                  [0, 0, 0, 1, 1, 0],
                  [0, 0, 0, 0, 1, 0],
                  [0, 0, 0, 0, 1, 1]]
        self.matrixTest(matrix, '>=')

    def testComparisonOperationLess(self):
        matrix = [[0, 0, 1, 0, 0, 0],
                  [0, 0, 1, 0, 0, 0],
                  [0, 0, 0, 0, 0, 0],
                  [0, 0, 0, 0, 0, 0],
                  [1, 1, 1, 1, 0, 1],
                  [0, 0, 0, 0, 0, 0]]
        self.matrixTest(matrix, '<')

    def testComparisonOperationGreater(self):
        matrix = [[0, 0, 0, 0, 1, 0],
                  [0, 0, 0, 0, 1, 0],
                  [1, 1, 0, 0, 1, 0],
                  [0, 0, 0, 0, 1, 0],
                  [0, 0, 0, 0, 0, 0],
                  [0, 0, 0, 0, 1, 0]]
        self.matrixTest(matrix, '>')

    def testList(self):
        self.assertEqual(self.quants[1] in self.quants, True)
        self.assertEqual(self.quants[0] in self.quants[1:], True)
        self.assertEqual(self.quants[2] in self.quants[3:], False)
        self.assertEqual(self.quants[1] in [], False)


if __name__ == "__main__":
    import sys
    if len(sys.argv) == 1:
        unittest.main()
    else:
        suite = unittest.TestLoader().loadTestsFromTestCase(eval(sys.argv[1:][0]))
        unittest.TextTestRunner().run(suite)



# -*- coding: utf-8 -*-

# Copyright (c) 2006-2010, Rectorate of the University of Freiburg
# Copyright (c) 2010, Andreas W. Liehr (liehr@users.sourceforge.net)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
This module provides a wrapper class that translates the
KnowledgeManager's summary dictionaries to an SQLite3 database.
"""

import sqlite3
import time
from pyphant.core.Helpers import (utf82uc, uc2utf8, emd52dict)
from pyphant.quantities import (Quantity,PhysicalUnit,_base_units)
PhysicalQuantity = Quantity
from types import (FloatType, IntType, LongType, StringTypes)

DBASE_VERSION = 2 #increment if there have been structural changes to the dbase!

def quantity2powers(quantity):
    numberOfBaseUnits = len(_base_units)
    if isinstance(quantity, Quantity):
        result = tuple(quantity.unit.powers)
        assert len(result) == numberOfBaseUnits, "Expecting %i base units, but got a tupple of %i unit powers insteat." % (numberOfBaseUnits,len(result))
        return result
    elif isinstance(quantity, (FloatType, IntType, LongType)):
        return (0, ) * numberOfBaseUnits
    else:
        raise ValueError("Expected (Quantity, FloatType, IntType, "\
                             "LongType) but got %s instead."\
                             % (type(quantity), ))

def str2number(str):
    try:
        value = int(str)
    except ValueError:
        try:
            value = long(str)
        except ValueError:
            value = float(str)
    return value

def quantity2dbase(quantity):
    if isinstance(quantity, (FloatType, IntType, LongType)):
        return quantity.__repr__()
    elif isinstance(quantity, Quantity):
        return "P%s;%s" % (quantity.value.__repr__(),
                            quantity.getUnitName())
    else:
        raise ValueError("Expected (Quantity, FloatType, IntType, "\
                             "LongType) but got %s instead."\
                             % (type(quantity), ))

def dbase2quantity(dbase):
    if isinstance(dbase, StringTypes):
        if dbase.startswith("P"):
            tmp = dbase[1:].split(';')
            return Quantity(str2number(tmp[0]), tmp[1])
        else:
            return str2number(dbase)
    else:
        raise ValueError("Broken FC unit in dbase: %s" % (dbase.__repr__(), ))

def unitname2latex(unitname):
    return unitname #TODO

def dbase2latex(dbase):
    if isinstance(dbase, StringTypes):
        if dbase.startswith("P"):
            tmp = dbase[1:].split(';')
            return unitname2latex(tmp[1])
        else:
            return dbase
    else:
        raise ValueError("Broken FC unit in dbase: %s" % (dbase.__repr__(), ))

def date2dbase(date):
    """extends a short datestring to YYYY-MM-DD_hh:mm:ss.ssssss standard
    """
    assert len(date) in [4, 7, 10, 13, 16, 19, 21, 22, 23, 24, 25, 26]
    date = date.replace(' ', '_')
    complete_str = '0000-01-01_00:00:00.000000'
    return date + complete_str[len(date):]

def emd52type(emd5):
    if emd5.endswith('d'):
        return 'fc'
    elif emd5.endswith('e'):
        return 'sc'
    else:
        raise ValueError(emd5)

def replace_type(str, type):
    if type == 'field':
        return str % ('fc', )
    elif type == 'sample':
        return str % ('sc', )

def get_wildcards(length, char, braces=False, commas=True):
    if braces:
        wc = '('
    else:
        wc = ''
    for index in xrange(length):
        wc += char
        if commas:
            wc += ','
        wc += ' '
    if commas:
        wc = wc[:-2]
    else:
        wc = wc[:-1]
    if braces:
        wc += ')'
    return wc

def create_table(table_name, columns, cursor, ignore_exists=False):
    if ignore_exists:
        query = "CREATE TABLE IF NOT EXISTS %s (" % (table_name, )
    else:
        query = "CREATE TABLE %s (" % (table_name, )
    for name, type in columns:
        query += name + " " + type + ", "
    query = query[:-2] + ")"
    cursor.execute(query)

def create_trigger(trigger_name, action, table_name,
                  statements, cursor):
    query = "CREATE TRIGGER %s AFTER %s ON %s "\
        "FOR EACH ROW BEGIN %s END"
    st_query = ''
    for st in statements:
        st_query += st + ';'
    cursor.execute(query % (trigger_name, action,
                            table_name, st_query))


class AnyValue():
    """Dummy class for use in search queries
    """
    def __init__(self):
        pass


class SQLiteWrapper(object):
    """Wrapper class for DC meta data <-> sqlite3
    """
    common_keys = ['longname', 'shortname', 'machine',
                   'creator', 'hash', 'date']
    writable_keys = ['storage']
    fast_keys = ['machine', 'creator', 'date', 'hash', 'type', 'id']
    all_keys = ['id', 'hash', 'longname', 'shortname', 'machine', 'creator',
                'date', 'type', 'attributes', 'storage', 'unit', 'columns',
                'dimensions']
    common_result_keys = common_keys + ['id', 'type', 'storage']
    one_to_one_search_keys = ['longname', 'shortname', 'machine',
                              'creator', 'hash', 'storage']
    one_to_one_result_keys = one_to_one_search_keys + ['date', 'id', 'type']
    common_search_keys = one_to_one_search_keys + \
                         ['id', 'attributes', 'date_from', 'date_to', 'col_of']
    fc_search_keys = common_search_keys + ['unit', 'dimensions',
                                           'dim_of', 'has_dim']
    sc_search_keys = common_search_keys + ['columns', 'has_col']
    sortable_keys = common_keys + ['storage', 'type']
    any_value = AnyValue()

    def __init__(self, database, timeout=60.0):
        """
        Arguments:
        - database: database to connect to
        """
        self.database = database
        self.timeout = timeout
        self.connection = None
        self.cursor = None

    def __enter__(self):
        assert self.connection == None
        assert self.cursor == None
        self.connection = sqlite3.connect(self.database, self.timeout,
                                          detect_types=sqlite3.PARSE_COLNAMES)
        self.cursor = self.connection.cursor()
        return self

    def __exit__(self, type, value, traceback):
        if type == None:
            try:
                self.connection.commit()
            except:
                print "Could not commit changes to database."
        if hasattr(self.cursor, 'close'):
            self.cursor.close()
        if hasattr(self.connection, 'close'):
            self.connection.close()
        self.cursor = None
        self.connection = None

    def __getitem__(self, emd5):
        if self.has_entry(emd5):
            if emd5.endswith('field'):
                return FCRowWrapper(emd5, self.cursor)
            elif emd5.endswith('sample'):
                return SCRowWrapper(emd5, self.cursor)
        raise KeyError(emd5)

    def setup_sqlite(self):
        sqlite3.register_converter('QUANTITY', dbase2quantity)
        sqlite3.register_converter('LATEX', dbase2latex)
        #clean tmp:
        self.cursor.execute("DELETE FROM km_temporary")

    def setup_dbase(self):
        #create tables:
        columns = [('sc_id', 'TEXT PRIMARY KEY UNIQUE NOT NULL'),
                   ('longname', 'TEXT'),
                   ('shortname', 'TEXT'),
                   ('machine', 'TEXT'),
                   ('creator', 'TEXT'),
                   ('date', 'TEXT'),
                   ('hash', 'TEXT'),
                   ('storage', 'TEXT')]
        create_table("km_sc", columns, self.cursor)
        columns[0] = ('fc_id', 'TEXT PRIMARY KEY UNIQUE NOT NULL')
        columns.insert(7, ('unit', 'TEXT'))
        columns.insert(8, ('bu_id', 'INT'))
        create_table("km_fc", columns, self.cursor)
        columns = [('sc_id', 'TEXT NOT NULL'),
                   ('fc_id', 'TEXT NOT NULL'),
                   ('fc_index', 'INT NOT NULL'),
                   ('', 'UNIQUE(sc_id, fc_id, fc_index)'),
                   ('', 'PRIMARY KEY(sc_id, fc_id, fc_index)')]
        create_table("km_sc_columns", columns, self.cursor)
        columns = [('fc_id', 'TEXT NOT NULL'),
                   ('dim_id', 'TEXT NOT NULL'),
                   ('dim_index', 'INT NOT NULL'),
                   ('', 'UNIQUE(fc_id, dim_id, dim_index)'),
                   ('', 'PRIMARY KEY(fc_id, dim_id, dim_index)')]
        create_table("km_fc_dimensions", columns, self.cursor)
        columns = [('dc_id', 'TEXT NOT NULL'),
                   ('key', 'TEXT NOT NULL'),
                   ('value', 'TEXT'),
                   ('', 'UNIQUE(dc_id, key)'),
                   ('', 'PRIMARY KEY(dc_id, key)')]
        create_table('km_attributes', columns, self.cursor)
        columns = [('dc_id', 'TEXT PRIMARY KEY UNIQUE NOT NULL')]
        create_table("km_temporary", columns, self.cursor)
        columns = [('bu_id', 'INTEGER PRIMARY KEY AUTOINCREMENT '\
                        'NOT NULL UNIQUE'),
                   ('m', 'INT'),
                   ('g', 'INT'),
                   ('s', 'INT'),
                   ('A', 'INT'),
                   ('K', 'INT'),
                   ('mol', 'INT'),
                   ('cd', 'INT'),
                   ('rad', 'INT'),
                   ('sr', 'INT'),
                   ('EUR', 'INT'),
                   ('bit', 'INT'),
                   ('', 'UNIQUE(m, g, s, A, K, mol, cd, rad, sr, EUR,bit)')]
        create_table('km_base_units', columns, self.cursor)
        columns = [('version', 'INT')]
        create_table('db_info', columns, self.cursor)
        self.cursor.execute("INSERT INTO db_info (version) VALUES (?)",
                            (DBASE_VERSION, ))
        #create triggers:
        create_trigger('trigger_del_fc', 'DELETE', 'km_fc',
                      ['DELETE FROM km_attributes WHERE dc_id=OLD.fc_id',
                       'DELETE FROM km_fc_dimensions WHERE fc_id=OLD.fc_id'],
                      self.cursor)
        create_trigger('trigger_del_sc', 'DELETE', 'km_sc',
                      ['DELETE FROM km_attributes WHERE dc_id=OLD.sc_id',
                       'DELETE FROM km_sc_columns WHERE sc_id=OLD.sc_id'],
                      self.cursor)
        create_trigger('trigger_del_tmp', 'DELETE', 'km_temporary',
                      ['DELETE FROM km_fc WHERE fc_id=OLD.dc_id',
                       'DELETE FROM km_sc WHERE sc_id=OLD.dc_id'],
                      self.cursor)
        self.setup_sqlite()

    def dbase_broken(self):
        try:
            self.cursor.execute("SELECT version FROM db_info")
            version = self.cursor.fetchone()[0]
            return version is not DBASE_VERSION
        except sqlite3.OperationalError:
            return True

    def has_entry(self, id):
        exe = self.cursor.execute
        if id == 'IndexMarker':
            type = 'fc'
        else:
            type = emd52type(id)
        exe("SELECT %s_id FROM km_%s WHERE %s_id=?" % (type, type, type),
            (id, ))
        return self.cursor.fetchone() != None

    def is_temporary(self, id):
        exe = self.cursor.execute
        exe("SELECT dc_id FROM km_temporary WHERE dc_id=?", (id, ))
        return self.cursor.fetchone() != None

    def _set_fc_keys(self, insert_dict, summary):
        exe = self.cursor.execute
        insert_dict['fc_id'] = summary['id']
        insert_dict['unit'] = quantity2dbase(summary['unit'])
        try:
            exe("INSERT OR ABORT INTO km_base_units "\
                    "(m, g, s, A, K, mol, cd, rad, sr, EUR, bit) "\
                    "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                quantity2powers(summary['unit']))
            l_row_id = self.cursor.lastrowid
        except sqlite3.IntegrityError:
            exe("SELECT bu_id FROM km_base_units WHERE m=? AND g=? "\
                    "AND s=? AND A=? AND K=? AND mol=? AND cd=? AND rad=? "\
                    "AND sr=? AND EUR=? AND bit=?", quantity2powers(summary['unit']))
            tmp = self.cursor.fetchone()
            assert tmp != None
            l_row_id = tmp[0]
        insert_dict['bu_id'] = l_row_id
        dimension_query = "INSERT INTO km_fc_dimensions VALUES (?, ?, ?)"
        for dim_id, dim_index in zip(summary['dimensions'],
                                     range(len(summary['dimensions']))):
            exe(dimension_query, (summary['id'], dim_id, dim_index))

    def set_entry(self, summary, storage, temporary=False):
        """Sets the meta data in the database according to the
        summary dictionary. If the according entry already exists,
        the database is not changed since the same emd5s should always
        reference the same (meta)data.
        Arguments:
        - summary: dictionary with meta data
        - storage: string type (e.g. path in local file system)
        - temporary: Flag that marks data to be deleted upon next
                     call to setup_dbase().
        """
        if self.has_entry(summary['id']):
            return
        exe = self.cursor.execute
        if temporary:
            exe("INSERT INTO km_temporary VALUES (?)",
                (summary['id'],))
        insert_dict = dict([(key, value) for key, value in \
                                summary.iteritems() if key in \
                                SQLiteWrapper.common_keys])
        insert_dict['storage'] = storage
        insert_dict['date'] = date2dbase(insert_dict['date'])
        type = emd52type(summary['id'])
        attr_query = "INSERT INTO km_attributes VALUES (?, ?, ?)"
        for key, value in summary['attributes'].iteritems():
            assert isinstance(key, StringTypes)
            if isinstance(value, StringTypes):
                value = utf82uc(value)
            exe(attr_query, (summary['id'], key, value.__repr__()))
        if type == 'fc':
            self._set_fc_keys(insert_dict, summary)
        else:
            insert_dict['sc_id'] = summary['id']
            column_query = "INSERT INTO km_sc_columns VALUES (?, ?, ?)"
            for fc_id, fc_index in zip(summary['columns'],
                                       range(len(summary['columns']))):
                exe(column_query, (summary['id'], fc_id, fc_index))
        insert_query = "INSERT INTO km_%s %s VALUES %s"
        value_list = []
        key_query = "("
        value_query = "("
        for key, value in insert_dict.iteritems():
            value_query += "?, "
            key_query += key + ", "
            value_list.append(value)
        key_query = key_query[:-2] + ")"
        value_query = value_query[:-2] + ")"
        insert_query = insert_query % (type, key_query, value_query)
        exe(insert_query, tuple(value_list))

    def set_temporary(self, entry_id, temporary):
        """Sets the given entry to temporary, which means it will be
        deleted upon restart or removes the temporary flag.
        - entry_id: emd5 of DataContainer
        - temporary: boolean"""
        if not self.has_entry(entry_id):
            return
        exe = self.cursor.execute
        if temporary:
            try:
                exe("INSERT OR ABORT INTO km_temporary VALUES (?)",
                    (entry_id,))
            except sqlite3.IntegrityError:
                pass # Already set to temporary!
        else:
            exe("DELETE FROM km_temporary WHERE dc_id=?",
                (entry_id,))

    def get_emd5_list(self):
        self.cursor.execute("SELECT fc_id FROM km_fc")
        emd5_list = self.cursor.fetchall()
        self.cursor.execute("SELECT sc_id FROM km_sc")
        emd5_list.extend(self.cursor.fetchall())
        return [row[0] for row in emd5_list]

    def verify_keys(self, keys, allowed):
        for key in keys:
            if not key in allowed:
                raise KeyError(key)

    def translate_result_key(self, key, type):
        if key == 'id':
            return replace_type("%s_id", type)
        elif key == 'type':
            return "'%s' AS type" % type
        elif key == 'unit':
            return 'unit AS "unit [QUANTITY]"'
        elif key == 'latex_unit':
            return 'unit AS "latex_unit [LATEX]"'
        else:
            return key

    def translate_unit_search(self, value):
        if isinstance(value, Quantity):
            value = value.unit.powers
        elif isinstance(value, (IntType, LongType, FloatType)):
            value = [0] * len(_base_units)
        elif isinstance(value, PhysicalUnit):
            value = value.powers
        else:
            raise ValueError(value)
        expr = '(bu_id IN (SELECT bu_id FROM km_base_units WHERE '\
            'm=? AND g=? AND s=? AND A=? AND K=? AND mol=? '\
            'AND cd=? AND rad=? AND sr=? AND EUR=? AND bit=?))'
        return (expr, value, True)

    def translate_attr_search(self, value, type):
        if value == {}:
            return ('1', [], True)
        expr = '('
        new_value = []
        for attr_key, attr_value in value.iteritems():
            if isinstance(attr_value, AnyValue):
                value_expr = ''
                new_value.append(attr_key)
            else:
                value_expr = ' AND value=?'
                if isinstance(attr_value, StringTypes):
                    attr_value = utf82uc(attr_value)
                new_value.extend([attr_key, attr_value.__repr__()])
            expr += '(%s IN (SELECT dc_id FROM km_attributes '\
                'WHERE key=?%s))' \
                % (replace_type('%s_id', type), value_expr)
            expr += ' AND '
        expr = expr[:-5] + ')'
        return (expr, new_value, True)

    def translate_list_search(self, key, value, type):
        id_str = replace_type('%s_id', type)
        if key == 'columns':
            table = 'km_sc_columns'
            index_str = 'fc_index'
            lid_str = 'fc_id'
        else:
            table = 'km_fc_dimensions'
            index_str = 'dim_index'
            lid_str = 'dim_id'
        qry = '(%s IN (SELECT %s FROM %s WHERE %s=? AND (%s IN (%s))))'\
            % (id_str, id_str, table, index_str, lid_str, '%s')
        new_value = []
        expr = '('
        for fc_search_dict, fc_index in zip(value, range(len(value))):
            fc_query, fc_values = self.get_andsearch_query(
                'field', ['id'], fc_search_dict, False)
            expr += qry % (fc_query, )
            new_value.append(fc_index)
            if fc_values != None:
                new_value.extend(fc_values)
            expr += ' AND '
        return (expr[:-5] + ')', new_value, True)

    def translate_parent_search(self, key, dc_search_dict, type):
        id_str = replace_type('%s_id', type)
        if dc_search_dict.has_key('type'):
            check = dc_search_dict.pop('type')
        else:
            check = None
        if key == 'col_of':
            assert check in [None, 'sample']
            parent_type = 'sample'
            parent_id_str = 'sc_id'
            child_id_str = 'fc_id'
            table = 'km_sc_columns'
        elif key == 'dim_of':
            assert check in [None, 'field']
            parent_type = 'field'
            parent_id_str = 'fc_id'
            child_id_str = 'dim_id'
            table = 'km_fc_dimensions'
        dc_query, dc_values = self.get_andsearch_query(
            parent_type, ['id'], dc_search_dict, False)
        expr = "(%s IN (SELECT %s FROM %s WHERE %s IN (%s)))" \
               % (id_str, child_id_str, table, parent_id_str, dc_query)
        return (expr, dc_values, True)

    def translate_child_search(self, key, child_search_dict):
        if key == 'has_col':
            if child_search_dict.has_key('type'):
                child_type = child_search_dict.pop('type')
                child_qry, child_values = self.get_andsearch_query(
                    child_type, ['id'], child_search_dict, False)
            else:
                fc_child_qry, fc_child_values = self.get_andsearch_query(
                    'field', ['id'], child_search_dict, False)
                sc_child_qry, sc_child_values = self.get_andsearch_query(
                    'sample', ['id'], child_search_dict, False)
                child_qry = "%s UNION ALL %s"
                child_qry = child_qry % (fc_child_qry, sc_child_qry)
                child_values = fc_child_values + sc_child_values
            id_str = 'sc_id'
            child_id_str = 'fc_id'
            table = 'km_sc_columns'
        elif key == 'has_dim':
            if child_search_dict.has_key('type'):
                assert child_search_dict.pop('type') == 'field'
            child_qry, child_values = self.get_andsearch_query(
                'field', ['id'], child_search_dict, False)
            id_str = 'fc_id'
            child_id_str = 'dim_id'
            table = 'km_fc_dimensions'
        expr = "(%s IN (SELECT %s FROM %s WHERE %s IN (%s)))" \
               % (id_str, id_str, table, child_id_str, child_qry)
        return (expr, child_values, True)

    def translate_search_dict(self, type, search_dict):
        where = ''
        values = []
        for key, value in search_dict.iteritems():
            extend = False
            if key in self.one_to_one_search_keys:
                expr = '%s=?' % key
            elif key == 'id':
                expr = '%s=?' % replace_type('%s_id', type)
            elif key == 'date_from':
                expr = 'date>=?'
                value = date2dbase(value)
            elif key == 'date_to':
                expr = 'date<?'
                value = date2dbase(value)
            elif key == 'unit':
                expr, value, extend = self.translate_unit_search(value)
            elif key == 'attributes':
                expr, value, extend = self.translate_attr_search(value, type)
            elif key == 'columns' or key == 'dimensions':
                expr, value, extend = self.translate_list_search(
                    key, value, type)
            elif key in ['dim_of', 'col_of']:
                expr, value, extend = self.translate_parent_search(
                    key, value, type)
            elif key in ['has_dim', 'has_col']:
                expr, value, extend = self.translate_child_search(
                    key, value)
            else:
                raise NotImplementedError(key)
            where += expr + " AND "
            if not extend:
                values.append(value)
            elif value is not None:
                values.extend(value)
        return where[:-5], values

    def get_andsearch_query(self, type, result_keys, search_dict, distinct):
        trans_res_keys = tuple([self.translate_result_key(key, type) \
                                    for key in result_keys])
        if type == 'field':
            table = 'km_fc'
        elif type == 'sample':
            table = 'km_sc'
        if search_dict == {}:
            qry = "SELECT%s %s FROM %s "
            values = None
        else:
            qry = "SELECT%s %s FROM %s WHERE "
        if distinct:
            dist_str = ' DISTINCT'
        else:
            dist_str = ' ALL'
        qry = (qry % (dist_str, get_wildcards(len(trans_res_keys),
                                              '%s'), table)) % trans_res_keys
        if search_dict != {}:
            where, values = self.translate_search_dict(type, search_dict)
            qry += where
        return qry, values

    def get_andsearch_result(self, result_keys, search_dict={},
                             order_by=None, order_asc=True,
                             limit=-1, offset=0, distinct=False):
        """returns a list of tuples filled with values of the result keys
        matching the constraints of search_dict.
        Arguments:
        - result_keys: List (of length >= 1) of keys to include in the
          result tuples.
        - search_dict: Dict mapping keys to constraint values.
          Use empty dict for no constraints at all
          possible keys: values (used relational operator[, type constraint]):
          'longname': str types (==)
          'shortname': str types (==)
          'machine': str types (==)
          'creator: str types (==)
          'date_from:' str types:
                       YYYY[-MM[-DD[_hh:[mm:[ss[.s[s[s[s[s[s]]]]]]]]]]] (>=)
          'date_to:' str types:
                     YYYY[-MM[-DD[_hh:[mm:[ss[.s[s[s[s[s[s]]]]]]]]]]] (<)
          'hash': str types (==)
          'id': str types: emd5 (==)
          'type': 'field' or 'sample' (==)
          'attributes': dict mapping attr. key to attr. value (==)
                        use (SQLiteWrapper instance).any_value
                        or (KM instance).any_value to skip value check
          'storage': str types (==)
          'unit': PhysicalUnit or number or Quantity (==, FC only)
          'dimensions': list of FC search dicts
                        (see above definitions, FC only)
          'dim_of': FC search dict (see above definitions, FC only)
          'col_of': SC search dict (see above definitions)
          'columns': list of FC search dicts (see above definitions, SC only)
          'has_col': DC search dict (see above defs, SC only)
          'had_dim': FC search dict (see above definitions, FC only)
        - order_by: element of result_keys to order the results by
                    or None for no special ordering
        - order_asc: whether to order ascending
        - limit: maximum number of results to return,
          set to -1 for no limit, default: -1
        - offset: number of search results to skip, default: 0
        - distinct: flag that indicates whether the result list
          should only contain distinct tuples.
        Usage Examples:
        Get list of all longnames:
           get_andsearch_result(['longname'], distinct=True)
           --> [('name1', ), ('name2', ), ...]
        Get id and shortname of all FCs that are parametrized by
        a time dimension along the primary axis:
           tunit = Quantity(1, 's')
           get_andsearch_result(['id', 'shortname'],
                                {'type':'field',
                                 'dimensions':[{'unit':tunit}]})
           --> [('emd5_1', 'name_1'), ('emd5_2', 'name_2'), ...]
        """
        if order_by == None:
            order = ''
        else:
            assert order_by in result_keys
            assert order_by in self.sortable_keys
            order = ' ORDER BY %s %s' \
                    % (order_by, {True:'ASC', False:'DESC'}[order_asc])
        assert isinstance(limit, int)
        assert isinstance(offset, int)
        if not search_dict.has_key('type'):
            self.verify_keys(result_keys, self.common_result_keys)
            self.verify_keys(search_dict.keys(), self.common_search_keys)
            fc_query, fc_values \
                = self.get_andsearch_query('field', result_keys,
                                           search_dict, distinct)
            sc_query, sc_values \
                = self.get_andsearch_query('sample', result_keys,
                                           search_dict, distinct)
            if distinct:
                dist_str = ''
            else:
                dist_str = ' ALL'
            query = "%s UNION%s %s%s LIMIT %d OFFSET %d"
            query = query % (fc_query, dist_str, sc_query, order, limit, offset)
            if search_dict != {}:
                values = fc_values + sc_values
            else:
                values = None
            mod_search_dict = search_dict
        else:
            if search_dict['type'] == 'field':
                 allowed_search_keys = self.fc_search_keys
                 allowed_result_keys = self.common_result_keys \
                                       + ['unit', 'latex_unit']
            elif search_dict['type'] == 'sample':
                 allowed_search_keys = self.sc_search_keys
                 allowed_result_keys = self.common_result_keys
            mod_search_dict = search_dict.copy()
            mod_search_dict.pop('type')
            self.verify_keys(mod_search_dict.keys(), allowed_search_keys)
            self.verify_keys(result_keys, allowed_result_keys)
            query, values = self.get_andsearch_query(
                search_dict['type'], result_keys, mod_search_dict, distinct)
            query = "%s%s LIMIT %d OFFSET %d" % (query, order, limit, offset)
        if values is not None and len(values) > 0:
            self.cursor.execute(query, values)
        else:
            self.cursor.execute(query)
        return self.cursor.fetchall()


class RowWrapper(object):

    def __init__(self, emd5, cursor, type):
        self.cursor = cursor
        self.emd5 = emd5
        self.emd5dict = emd52dict(emd5)
        self.emd5dict['id'] = emd5
        self.select_query = "SELECT %s FROM km_%s WHERE %s_id=?"\
            % ('%s', type, type)
        self.update_query = "UPDATE km_%s SET %s=? WHERE %s_id=?"\
            % (type, '%s', type)
        self.attr_query = "SELECT key, value FROM km_attributes WHERE dc_id=?"

    def __getitem__(self, key):
        if not key in SQLiteWrapper.all_keys:
            raise ValueError("'%s' is not a valid key!" % (key, ))
        elif key in SQLiteWrapper.fast_keys:
            return self.emd5dict[key]
        elif key == 'attributes':
            self.cursor.execute(self.attr_query, (self.emd5, ))
            # Eval is evil! TODO: Nail possible types of avalue and write
            # according wrappers.
            return dict([(akey, eval(avalue)) for akey, avalue in self.cursor])
        else:
            self.cursor.execute(self.select_query % (key, ), (self.emd5, ))
            return self.cursor.fetchone()[0]

    def __setitem__(self, key, value):
        if key not in SQLiteWrapper.writable_keys:
            raise ValueError("'%s' is read only!" % (key, ))
        else:
            self.cursor.execute(self.update_query % (key, ),
                                (value, self.emd5))


class FCRowWrapper(RowWrapper):

    def __init__(self, emd5, cursor):
        RowWrapper.__init__(self, emd5, cursor, 'fc')
        self.dimension_query = "SELECT dim_id FROM km_fc_dimensions "\
            "WHERE fc_id=? ORDER BY dim_index ASC"
        self.unit_query = 'SELECT unit AS "unit [%s]" FROM km_fc WHERE fc_id=?'

    def __getitem__(self, key):
        if key == 'dimensions':
            self.cursor.execute(self.dimension_query, (self.emd5, ))
            dimensions = [row[0] for row in self.cursor]
            if dimensions == []:
                dimensions = [u'IndexMarker']
            return dimensions
        elif key == 'unit':
            self.cursor.execute(self.unit_query % 'QUANTITY', (self.emd5, ))
            return self.cursor.fetchone()[0]
        elif key == 'latex_unit':
            self.cursor.execute(self.unit_query % 'LATEX', (self.emd5, ))
            return self.cursor.fetchone()[0]
        elif key == 'columns':
            raise KeyError(key)
        else:
            return RowWrapper.__getitem__(self, key)


class SCRowWrapper(RowWrapper):

    def __init__(self, emd5, cursor):
        RowWrapper.__init__(self, emd5, cursor, 'sc')
        self.column_query = "SELECT fc_id FROM km_sc_columns "\
            "WHERE sc_id=? ORDER BY fc_index ASC"

    def __getitem__(self, key):
        if key == 'columns':
            self.cursor.execute(self.column_query, (self.emd5, ))
            return [row[0] for row in self.cursor]
        elif key == 'unit' or key == 'dimensions':
            raise KeyError(key)
        else:
            return RowWrapper.__getitem__(self, key)



# -*- coding: utf-8 -*-

# Copyright (c) 2006-2009, Rectorate of the University of Freiburg
# Copyright (c) 2009, Andreas W. Liehr (liehr@users.sourceforge.net)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

u"""Provides unittest classes FieldContainerTestCase, SampleContainerTest.
"""


import unittest
import scipy
import copy
from pyphant.quantities import Quantity
from pyphant.core.DataContainer import (INDEX,
                                        generateIndex,
                                        FieldContainer,
                                        SampleContainer,
                                        DataContainer,
                                        assertEqual)
import numpy.testing as nt
import numpy

class DataContainerTestCase(unittest.TestCase):
    def setUp(self):
        self.dc = DataContainer('variable', 'v')
        self.dc0 = DataContainer('variable', 'v_0')

    def testNoSubscriptPersistent(self):
        result = self.dc.appendSubscript('i')
        self.assertEqual(result, 'v_{i}')
        self.assertEqual(self.dc.shortname, 'v_{i}')

    def testSubscriptPersistent(self):
        result = self.dc0.appendSubscript('i')
        self.assertEqual(result, 'v_{0,i}')
        self.assertEqual(self.dc0.shortname, 'v_{0,i}')

    def testNoSubscriptNotPersistent(self):
        result = self.dc.appendSubscript('i', persistent=False)
        self.assertEqual(result, 'v_{i}')
        self.assertEqual(self.dc.shortname, 'v')

    def testSubscriptNotPersistent(self):
        result = self.dc0.appendSubscript('i', persistent=False)
        self.assertEqual(result, 'v_{0,i}')
        self.assertEqual(self.dc0.shortname, 'v_0')

    def testDefaultAttribute(self):
        """
        Checking if the default attribute has correctly been set to an empty
        dictionary.
        """
        self.assertEqual(self.dc.attributes,
                         {},
                         'Expected an empty dictionary as attribute, \
but found %s.' % self.dc.attributes )

class FieldContainerTestCase(unittest.TestCase):
    def setUp(self):
        self.testData = scipy.array([[0., 1., 2.], [3., 4., 5.], [6., 7., 8.]])
        self.testMask = self.testData > 5
        self.longname = u"Sampled Data"
        self.shortname = u"I\\omega"
        self.unit = Quantity('3.14 m')

    def testGenerateIndex(self):
        i=0
        n=1000
        index = generateIndex(i,n)
        self.assertEqual(index.dimensions, INDEX)
        self.assertEqual(index.longname, 'Index')
        self.assertEqual(index.shortname, 'i')
        self.assertEqual(index.unit, 1)
        self.assertFalse(index.error)
        self.assertTrue(scipy.alltrue(index.data == scipy.arange(0,n)))

    def testLabeling(self):
        field = FieldContainer(self.testData,
                               1,
                               longname=self.longname,
                               shortname=self.shortname)
        self.assertEqual(field.label,"%s $%s(i,j)$ / a.u." % (self.longname,
                                                              self.shortname))
        field = FieldContainer(self.testData,
                               self.unit,
                               longname=self.longname,
                               shortname=self.shortname)
        self.assertEqual(field.label,"%s $%s(i,j)$ / %s" % (self.longname,
                                                            self.shortname,
                                                            self.unit))

    def testSeal(self):
        field = FieldContainer(self.testData,
                               1,
                               longname=self.longname,
                               shortname=self.shortname)
        field.seal()
        self.assertNotEqual(field.id, None)
        self.assertNotEqual(field.hash, None)
        try:
            field.data = scipy.array([1, 2, 3])
        except TypeError, e:
            pass
        else:
            self.fail("Modification of sealed FieldContainer was not \
                        prohibited.")
        try:
            field.data[1] = 4
        except (RuntimeError, ValueError):
            pass
        else:
            self.fail("Modification of sealed FieldContainer was not \
                        prohibited.")
        try:
            field.dimensions[0] = copy.deepcopy(field)
        except TypeError, e:
            pass
        else:
            self.fail("Modification of sealed FieldContainer's dimension \
                        was not prohibited.")

#This test is broken since it produces an invalid FieldContainer.
#I am not sure how to fix it or what its intent is.
#FIXME:
#    def testSealedFieldContainerWithDirectDimensionsModification(self):
#        field = FieldContainer(self.testData,1,longname=self.longname,shortname=self.shortname)
#        field.dimensions=[copy.deepcopy(field)]
#        try:
#            field.seal()
#        except AttributeError, e:
#            self.fail("Direct setting of attribute .dimensions has not instantiated a dimensionList object: %s"%e)

    def testDeepcopy(self):
        field = FieldContainer(self.testData,
                               1,
                               longname=self.longname,
                               shortname=self.shortname)
        copiedField=copy.deepcopy(field)
        self.assertEqual(field, copiedField)
        field.seal()
        copiedField.seal()
        self.assertEqual(field, copiedField)
        #equal because only real data is considered:
        self.assertEqual(field.hash, copiedField.hash)
        #unique due to timestamp in dimension ids:
        self.assertNotEqual(field.id, copiedField.id)

    def testEqual(self):
        field1 = FieldContainer(self.testData,
                                1,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                1,
                                longname=self.longname,
                                shortname=self.shortname)
        self.assertEqual(field1, field2)
        field1 = FieldContainer(self.testData,
                                1,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                1,
                                longname=self.longname + "bu",
                                shortname=self.shortname)
        self.assertEqual(field1, field2)
        field1 = FieldContainer(self.testData,
                                1,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                1,
                                longname=self.longname,
                                shortname=self.shortname + 'a')
        self.assertEqual(field1, field2)
        field1 = FieldContainer(self.testData,
                                "1 km",
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                "1000 m",
                                longname=self.longname,
                                shortname=self.shortname)
        self.assertEqual(field1, field2)
        #necessary to test the != operator (__ne__):
        assert(not (field1 != field2))
        field1 = FieldContainer(self.testData,
                                "1 km",
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                longname=self.longname,
                                shortname=self.shortname)
        self.assertNotEqual(field1, field2)
        field1 = FieldContainer(self.testData,
                                1,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                "1 m",
                                longname=self.longname,
                                shortname=self.shortname)
        self.assertNotEqual(field1, field2)
        field1 = FieldContainer(self.testData,
                                1,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                "1 m",
                                longname=self.longname + "bu",
                                shortname=self.shortname)
        self.assertNotEqual(field1, field2)
        #Test inequality of attributes
        field1 = FieldContainer(self.testData,
                                1,
                                longname=self.longname,
                                shortname=self.shortname,
                                attributes={'set':True})
        field2 = FieldContainer(self.testData,
                                1,
                                longname=self.longname,
                                shortname=self.shortname)
        self.assertNotEqual(field1, field2)
        #Test inequality of dimensions
        field1 = FieldContainer(self.testData,
                                1,longname=self.longname,
                                shortname=self.shortname)
        field1.dimensions[0].data *= 10
        field2 = FieldContainer(self.testData,
                                1,
                                longname=self.longname,
                                shortname=self.shortname)
        self.assertNotEqual(field1, field2)

    def testEqualMasked(self):
        field1 = FieldContainer(self.testData,
                                1,
                                mask=self.testMask,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                1,mask=self.testMask,
                                longname=self.longname,
                                shortname=self.shortname)
        self.assertEqual(field1, field2)
        field1 = FieldContainer(self.testData,
                                1,
                                mask=self.testMask,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                1,
                                mask=self.testMask,
                                longname=self.longname + "bu",
                                shortname=self.shortname)
        self.assertEqual(field1, field2)
        field1 = FieldContainer(self.testData,
                                "1 km",
                                mask=self.testMask,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                "1000 m",
                                mask=self.testMask,
                                longname=self.longname,
                                shortname=self.shortname)
        self.assertEqual(field1, field2)
        #necessary to test the != operator (__ne__)
        assert(not (field1 != field2))
        testMask2 = self.testMask.copy()
        testMask2[0, 2] = not self.testMask[0, 2]
        field1 = FieldContainer(self.testData,
                                1,
                                mask=self.testMask,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                1,
                                mask=testMask2,
                                longname=self.longname,
                                shortname=self.shortname)
        self.assertNotEqual(field1, field2)
        field1 = FieldContainer(self.testData,
                                1,
                                mask=self.testMask,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                1,
                                mask=testMask2,
                                longname=self.longname + "bu",
                                shortname=self.shortname)
        self.assertNotEqual(field1, field2)
        field1 = FieldContainer(self.testData,
                                "1 km",
                                mask=self.testMask,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                "1000 m",
                                mask=testMask2,
                                longname=self.longname,
                                shortname=self.shortname)
        self.assertNotEqual(field1, field2)
        assert(field1 != field2) #necessary to test the != operator (__ne__)
        field1 = FieldContainer(self.testData,
                                "1 km",
                                mask=self.testMask,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                mask=self.testMask,
                                longname=self.longname,
                                shortname=self.shortname)
        self.assertNotEqual(field1, field2)
        field1 = FieldContainer(self.testData,
                                1,
                                mask=self.testMask,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                "1 m",
                                mask=self.testMask,
                                longname=self.longname,
                                shortname=self.shortname)
        self.assertNotEqual(field1, field2)
        field1 = FieldContainer(self.testData,
                                1,
                                mask=self.testMask,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                "1 m",
                                mask=self.testMask,
                                longname=self.longname + "bu",
                                shortname=self.shortname)
        self.assertNotEqual(field1, field2)

    def testAddition(self):
        field1 = FieldContainer(self.testData,
                                1,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                1,
                                longname=self.longname,
                                shortname=self.shortname)
        sumField = field1 + field2
        nt.assert_array_almost_equal(sumField.data, field1.data + field2.data)
        self.assertEqual(1, sumField.unit)
        self.assertEqual(sumField.shortname,
                         u"%s + %s" % (self.shortname, self.shortname))
        field1 = FieldContainer(self.testData,
                                '1 m',
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                '2 m',
                                longname=self.longname,
                                shortname=self.shortname)
        sumField = field1 + field2
        nt.assert_array_almost_equal(sumField.data, field1.data / 2+field2.data)
        self.assertEqual(sumField.unit, Quantity('2 m'))
        self.assertEqual(sumField.shortname, u"%s + %s" \
                             % (self.shortname, self.shortname))
        sumField = field2 + field1
        nt.assert_array_almost_equal(sumField.data,
                                     field1.data / 2 + field2.data)
        self.assertEqual(sumField.unit, Quantity('2 m'))
        self.assertEqual(sumField.shortname,
                         u"%s + %s" % (self.shortname, self.shortname))
        field1 = FieldContainer(self.testData,
                                '1 mm',
                                longname=self.longname,
                                shortname=self.shortname)
        sumField = field1 + field2
        nt.assert_array_almost_equal(sumField.data,
                                     field1.data / 2000 + field2.data)
        self.assertEqual(sumField.unit, Quantity('2 m'))
        self.assertEqual(sumField.shortname,
                         u"%s + %s" % (self.shortname, self.shortname))

    def testSubtraction(self):
        field1 = FieldContainer(numpy.random.randn(7, 13),
                                1,
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(numpy.random.randn(7, 13),
                                1,
                                longname=self.longname,
                                shortname=self.shortname)
        sumField = field1 - field2
        nt.assert_array_almost_equal(sumField.data, field1.data - field2.data)
        self.assertEqual(1, sumField.unit)
        self.assertEqual(sumField.shortname,
                         u"%s - %s" % (self.shortname, self.shortname))
        field1 = FieldContainer(self.testData,
                                '1 m',
                                longname=self.longname,
                                shortname=self.shortname)
        field2 = FieldContainer(self.testData,
                                '2 m',
                                longname=self.longname,
                                shortname=self.shortname)
        sumField = field1 - field2
        nt.assert_array_almost_equal(sumField.data,
                                     field1.data / 2 - field2.data)
        self.assertEqual(sumField.unit, Quantity('2 m'))
        self.assertEqual(sumField.shortname,
                         u"%s - %s" % (self.shortname, self.shortname))
        sumField = field2 - field1
        nt.assert_array_almost_equal(sumField.data,
                                     field2.data - field1.data / 2)
        self.assertEqual(sumField.unit, Quantity('2 m'))
        self.assertEqual(sumField.shortname,
                         u"%s - %s" % (self.shortname, self.shortname))
        field1 = FieldContainer(self.testData,
                                '1 mm',
                                longname=self.longname,
                                shortname=self.shortname)
        sumField = field1 - field2
        nt.assert_array_almost_equal(sumField.data,
                                     field1.data / 2000 - field2.data)
        self.assertEqual(sumField.unit, Quantity('2 m'))
        self.assertEqual(sumField.shortname,
                         u"%s - %s" % (self.shortname, self.shortname))

class IsValidFieldContainer(unittest.TestCase):
    def setUp(self):
        self.field = FieldContainer(numpy.random.randn(7, 13),
                                    longname="voltage",
                                    shortname="U",
                                    unit="1V")

    def testWrongDimension(self):
        self.field.dimensions[0].data = self.field.dimensions[0].data[:-1]
        self.assertFalse(self.field.isValid())

    def testWrongDimensionNumber(self):
        self.field.dimensions.append(copy.deepcopy(self.field.dimensions[0]))
        self.assertFalse(self.field.isValid())
        self.field.dimensions = [self.field.dimensions[0]]
        self.assertFalse(self.field.isValid())

    def testWrongMask(self):
        shape = list(self.field.data.shape)
        self.field.mask = numpy.ones(shape)
        self.assertTrue(self.field.isValid())
        shape[0] = shape[0] + 1
        self.field.mask = numpy.ones(shape)
        self.assertFalse(self.field.isValid())

    def testWrongError(self):
        shape = list(self.field.data.shape)
        self.field.error = numpy.zeros(shape)
        self.assertTrue(self.field.isValid())
        shape[0] = shape[0] + 1
        self.field.error = numpy.ones(shape)
        self.assertFalse(self.field.isValid())

    def testDimension0HasSameShapeAsField(self):
        self.field.dimensions[0] = copy.deepcopy(self.field)
        self.assertFalse(self.field.isValid())

    def testDimension1HasSameShapeAsField(self):
        self.field.dimensions[1] = copy.deepcopy(self.field)
        self.assertFalse(self.field.isValid())


class SampleContainerTest(unittest.TestCase):
    def setUp(self):
        self.rows = 100
        self.intSample = FieldContainer(scipy.arange(0, self.rows),
                                        Quantity('1m'),
                                        #dimensions=INDEX,
                                        longname=u"Integer sample",
                                        shortname=u"i")
        self.intSample2 = FieldContainer(2 * scipy.arange(0, self.rows),
                                        Quantity('1m'),
                                        longname=u"Integer sample No2",
                                        shortname=u"i2")
        self.floatSample = FieldContainer(scipy.arange(self.rows / 2,
                                                       self.rows,
                                                       0.5),
                                          Quantity('1s'),
                                          longname=u"Float sample",
                                          shortname=u"t")
        self.shortFloatSample1 = FieldContainer(scipy.array([1., 2., 10.]),
                                          Quantity('1.0 s**2'),
                                          longname=u"Short Float sample 1",
                                          shortname=u"short1")
        self.shortFloatSample2 = FieldContainer(scipy.array([5., 10., 1.]),
                                          Quantity('1.0 kg * m'),
                                          longname=u"Short Float sample 2",
                                          shortname=u"short2")
        self.desc = scipy.dtype({'names':[u'i', u't'],
                                 'formats':[self.intSample.data.dtype,
                                            self.floatSample.data.dtype],
                                 'titles':[self.intSample.longname,
                                           self.floatSample.longname]})
        self.data = scipy.rec.fromarrays([self.intSample.data,
                                        self.floatSample.data],
                                       dtype=self.desc)
        self.longname = u"Toller Sample"
        self.shortname = u"phi"
        self.sampleContainer = SampleContainer([self.intSample,
                                                self.floatSample],
                                               self.longname,
                                               self.shortname)
        self.sampleContainerNeu = SampleContainer([self.intSample,
                                                   self.intSample2,
                                                self.floatSample,
                                                self.shortFloatSample1,
                                                self.shortFloatSample2],
                                               "New Sample",
                                               "NewSC")

    def runTest(self):
        return


class AlgebraSampleContainerTests(SampleContainerTest):

    def testNodeTransformer(self):
        from pyphant.core.AstTransformers import (ReplaceName, ReplaceOperator,
                                                  ReplaceCompare)
        rpn = ReplaceName(self.sampleContainer)
        import ast
        exprStr = 'col("i") / (col("t") + col("t"))'
        expr = compile(exprStr, "<TestCase>", 'eval', ast.PyCF_ONLY_AST)
        replacedExpr = rpn.visit(expr)
        print rpn.localDict
        print ast.dump(replacedExpr)
        rpc = ReplaceCompare(rpn.localDict)
        factorExpr = rpc.visit(replacedExpr)
        rpo = ReplaceOperator(rpn.localDict)
        factorExpr = rpo.visit(factorExpr)
        print ast.dump(factorExpr)

    def testCalcColumn(self):
        exprStr = 'col("i") / (col("t") + col("t")) + "1km/s"'
        column = self.sampleContainer.calcColumn(exprStr, 'v', 'velocity')
        #print self.sampleContainerNeu['i']
        #print self.sampleContainerNeu['t']
        #print self.sampleContainerNeu['i2']
        #print column
        exprStr = 'col("i") / (col("t") + col("t")) + "1km"'
        self.assertRaises(ValueError, self.sampleContainer.calcColumn,
                          exprStr, 'v', 'velocity')

    def testCalcColumnExplicit(self):
        exprStr = 'col("short1") * col("short2") - "10 kg * m * s**2"'
        columnOut = self.sampleContainerNeu.calcColumn(
            exprStr, 'Test1', 'Mult und Minus')
        #print(columnOut)
        columnCheck = FieldContainer(scipy.array([-5., 10., 0.]),
                                     unit='1 kg * m * s**2',
                                     shortname='Test1Check',
                                     longname='Test1Checken')
        #print(columnCheck.unit)
        #print(type(columnCheck.unit))
        checkIfDataEqual = columnOut.data == columnCheck.data
        if checkIfDataEqual.all() and columnOut.unit == columnCheck.unit:
            print('Explicit check No1 ok.')
        else:
            raise ValueError
        exprStr = '(col("short1") / col("short2") - "10.0 s**2/(kg*m)") ' + \
                  '> ("-5.0 s**2/(kg*m)")'
        columnOut = self.sampleContainerNeu.calcColumn(
            exprStr, 'Test1', 'Mult und Minus')
        #print(columnOut)
        #print(columnOut.unit)
        columnCheck = FieldContainer(scipy.array([False, False, True]),
                                     shortname='Test2Check',
                                     longname='Test2Checken')
        checkIfDataEqual = columnOut.data == columnCheck.data
        if checkIfDataEqual.all() and columnOut.unit == columnCheck.unit:
            print('Explicit check No2 ok.')
        else:
            raise ValueError
        exprStr = 'col("short1") * col("short2") - "10 kg * s**2"'
        self.assertRaises(
            ValueError, self.sampleContainerNeu.calcColumn, exprStr,
            'Test1assert', 'Test1AssertRaise')
        exprStr = '"1.0m" / "0.0m"'
        self.assertRaises(ZeroDivisionError,
                          self.sampleContainerNeu.calcColumn, exprStr,\
                                            'TestZero', 'Division by Zero')

    def testAlgebraPlus(self):
        from pyphant.core.DataContainer import FieldContainer
        expr = 'col("i") + col("i2") + "-10 m"'
        columnOut = self.sampleContainerNeu.calcColumn(
            expr, 'OutPlus', 'OutcomePlus')
        columnCheck = FieldContainer(3 * scipy.arange(0, self.rows) - 10,
                                unit='1 m',
                                longname="Outcome Check",
                                shortname="OC")
        checkIfDataEqual = columnOut.data == columnCheck.data
        if checkIfDataEqual.all() and columnOut.unit == columnCheck.unit:
            print("Adding works.")
        else:
            raise ValueError

    def testAlgebraMinus(self):
        from pyphant.core.DataContainer import FieldContainer
        expr = 'col("i") - col("i2") - "-10 km"'
        columnOut = self.sampleContainerNeu.calcColumn(
            expr, 'OutMinus', 'OutcomeMinus')
        columnCheck = FieldContainer(-1 * scipy.arange(0, self.rows) + 10000,
                                unit='1 m',
                                longname="Outcome Check",
                                shortname="OC")
        checkIfDataEqual = columnOut.data == columnCheck.data
        if checkIfDataEqual.all() and columnOut.unit == columnCheck.unit:
            print("Subtracting works.")
        else:
            raise ValueError

    def testAlgebraMult(self):
        from pyphant.core.DataContainer import FieldContainer
        expr = 'col("i") * col("i") * 0.5'
        columnOut = self.sampleContainerNeu.calcColumn(
            expr, 'OutMult', 'OutcomeMult')
        columnCheckData = scipy.array(
            map(lambda x: x ** 2 * 0.5, scipy.arange(0, self.rows)))
        columnCheck = FieldContainer(columnCheckData,
                                unit='1 m**2',
                                longname="Outcome Check",
                                shortname="OC")
        checkIfDataEqual = columnOut.data == columnCheck.data
        if checkIfDataEqual.all() and columnOut.unit == columnCheck.unit:
            print("Multiplying works.")
        else:
            raise ValueError

    def testAlgebraDiv(self):
        from pyphant.core.DataContainer import FieldContainer
        expr = 'col("i") / 2.'
        expr2 = '(col("i") + "1 m") / (col("i") + "1 m")'
        columnOut = self.sampleContainerNeu.calcColumn(
            expr, 'OutDiv', 'OutcomeDiv')
        #print(columnOut)
        columnOut2 = self.sampleContainerNeu.calcColumn(
            expr2, 'OutDiv2', 'OutcomeDiv2')
        #print(columnOut.unit)
        #columnCheckData = scipy.array([1 for i in range(0,self.rows)])
        columnCheckData = 0.5 * scipy.arange(0., float(self.rows))
        columnCheckData2 = scipy.array([1. for i in range(0, self.rows)])
        columnCheck = FieldContainer(columnCheckData,
                                unit='1 m',
                                longname="Outcome Check",
                                shortname="OC")
        columnCheck2 = FieldContainer(columnCheckData2,
                                unit='1.0',
                                longname="Outcome Check",
                                shortname="OC")
        checkIfDataEqual = columnOut.data == columnCheck.data
        checkIfDataEqual2 = columnOut2.data == columnCheck2.data
        if checkIfDataEqual.all() and checkIfDataEqual2.all() \
               and columnOut.unit == columnCheck.unit:
            print("Dividing works.")
        else:
            raise ValueError

    def testDimensions(self):
        fc1 = FieldContainer(numpy.array([4, 5, 6]), unit=Quantity('1 m'),
                             dimensions=[FieldContainer(
                                 numpy.array([1, 2, 3]), unit=Quantity('1 s'))],
                             shortname='s1')
        fc2 = FieldContainer(numpy.array([1, 2, 3]), unit=Quantity('1 km'),
                             dimensions=[FieldContainer(
                                 numpy.array([10, 20, 30]),
                                 unit=Quantity('0.1 s'))],
                             shortname='s2')
        sc = SampleContainer(columns=[fc1, fc2])
        expr = "col('s1') + col('s2')"
        dims = sc.calcColumn(expr, 'Add', 'a').dimensions
        self.assertEqual(dims, fc1.dimensions)
        self.assertEqual(dims, fc2.dimensions)
        fc2.dimensions[0].unit = Quantity('0.11 s')
        self.assertRaises(ValueError, sc.calcColumn, expr, 'Add', 'a')


class CommonSampleContainerTests(SampleContainerTest):
    def testLabeling(self):
        self.assertEqual(self.sampleContainer.label,
                         "%s %s" % (self.longname,self.shortname))

    def testDeepcopy(self):
        sample = self.sampleContainer
        copiedSample = copy.deepcopy(sample)
        self.assertEqual(sample, copiedSample)
        sample.seal()
        copiedSample.seal()
        self.assertEqual(sample, copiedSample)
        #equal because only real data is considered:
        self.assertEqual(sample.hash, copiedSample.hash)
        #unique due to timestamp in dimension ids:
        self.assertNotEqual(sample.id, copiedSample.id)

    def testSeal(self):
        sample = self.sampleContainer
        sample.seal()
        self.assertNotEqual(sample.id, None)
        self.assertNotEqual(sample.hash, None)
        try:
            sample.data=scipy.array([1, 2, 3])
        except TypeError, e:
            pass
        else:
            self.fail("Modification of sealed FieldContainer was not "
                      "prohibited.")
        try:
            sample['i'].data[0] = 4
        except (RuntimeError, ValueError):
            pass
        else:
            self.fail("Modification of sealed FieldContainer was not "
                      "prohibited.")

    def testSingleSample(self):
        #string = numpy.rec.fromrecords([(s,) for s in [u'Hello',u'World!',u'Bäh!']])
#        strings =
        uField = FieldContainer(scipy.array([u'Hello', u'World!', u'Bäh!']),
                                longname=u'Text',
                                shortname='\gamma')
        #was: table  = SampleContainer([uField]). Intent unclear:
        table = SampleContainer([uField])

    def testStringSample(self):
        #string = numpy.rec.fromrecords([(s,) for s in [u'Hello',u'World!',u'Bäh!']])
#        strings =
        uField = FieldContainer(scipy.array([u'Hello', u'World!', u'Bäh!']),
                                longname=u'Text',
                                shortname='\gamma')
        sample = SampleContainer([uField])
        sample.seal()
        #print sample,sample.data

    def testEqual(self):
        #Test Equality
        sample1 = SampleContainer([self.intSample, self.floatSample],
                                  longname=self.longname,
                                  shortname=self.shortname)
        sample2 = SampleContainer([self.intSample, self.floatSample],
                                  longname=self.longname,
                                  shortname=self.shortname)
        self.assertEqual(sample1, sample2)
        #Test inequality due to different longnames
        sample1 = SampleContainer([self.intSample, self.floatSample],
                                  longname=self.longname,
                                  shortname=self.shortname)
        sample2 = SampleContainer([self.intSample, self.floatSample],
                                  longname=self.longname + 'bu',
                                  shortname=self.shortname)
        self.assertNotEqual(sample1, sample2)
        #Test inequality to due different shortnames
        sample1 = SampleContainer([self.intSample, self.floatSample],
                                  longname=self.longname,
                                  shortname=self.shortname)
        sample2 = SampleContainer([self.intSample, self.floatSample],
                                  longname=self.longname,
                                  shortname=self.shortname + 'a')
        self.assertNotEqual(sample1, sample2)
        #Test inequality of attributes
        sample1 = SampleContainer([self.intSample, self.floatSample],
                                  longname=self.longname,
                                  shortname=self.shortname)
        sample2 = SampleContainer([self.intSample, self.floatSample],
                                  longname=self.longname,
                                  shortname=self.shortname,
                                  attributes={'set':True})
        self.assertNotEqual(sample1, sample2)


class SampleContainerSlicingTests(SampleContainerTest):
    def setUp(self):
        SampleContainerTest.setUp(self)
        time_data = numpy.array([10.0, 20.0, 30.0, 5.0, 9000.0])
        time_error = numpy.array([1.0, 2.0, 3.0, .5, 900.0])
        time_unit = Quantity('2s')
        time_FC = FieldContainer(time_data, time_unit, time_error,
                                 None, None,
                                 "Zeit", "t",
                                 None, False)
        length_data = numpy.array([-20.0, 0.0, 20.0, 10.0, 5.5])
        length_error = numpy.array([2.0, 0.1, 2.0, 1.0, .5])
        length_unit = Quantity('1000m')
        length_FC = FieldContainer(length_data, length_unit, length_error,
                                   None, None,
                                   "Strecke", "l",
                                   None, False)
        temperature_data = numpy.array([[10.1, 10.2, 10.3],
                                        [20.1, 20.2, 20.3],
                                        [30.1, 30.2, 30.3],
                                        [40.1, 40.2, 40.3],
                                        [50.1, 50.2, 50.3]])
        temperature_error = numpy.array([[0.1, 0.2, 0.3],
                                         [1.1, 1.2, 1.3],
                                         [2.1, 2.2, 2.3],
                                         [3.1, 3.2, 3.3],
                                         [4.1, 4.2, 4.3]])
        temperature_unit = Quantity('1mK')
        temperature_FC = FieldContainer(temperature_data,
                                        temperature_unit,
                                        temperature_error,
                                        None, None,
                                        "Temperatur", "T",
                                        None, False)
        self.sc2d = SampleContainer([length_FC, temperature_FC, time_FC],
                                    "Test Container", "TestC")
        self.sc2d["t"].dimensions[0].unit = Quantity('1m')
        self.sc2d["t"].dimensions[0].data = numpy.array([-20, -10, 0, 10, 20])
        self.sc2d["l"].dimensions[0].unit = Quantity('5m')
        self.sc2d["l"].dimensions[0].data = numpy.array([-4, -2, 0, 2, 4])
        self.sc2d["T"].dimensions[0].unit = Quantity('2m')
        self.sc2d["T"].dimensions[0].data = numpy.array([-10, -5, 0, 5, 10])
        self.sc2d["T"].dimensions[1].unit = Quantity('10nm')
        self.sc2d["T"].dimensions[1].data = numpy.array([-1, 0, 1])

    #purely one dimensional Tests:
    def testConsistancy(self):
        result1 = self.sampleContainer.filter(
            '"20m" < col("i") and "80m" > col("i")')
        result2 = self.sampleContainer.filter('"20m" < col("i") < "80m"')
        self.assertEqual(result1[0], result2[0])
        self.assertEqual(result1[1], result2[1])

    def testSimpleUnicodeExpression(self):
        result = self.sampleContainer.filter(u'"50m" <= col("i") < "57m"')
        self.assertEqual(len(result.columns), 2)
        self.assertEqual(len(result[0].data), 7)
        self.assertEqual(len(result[1].data), 7)
        expected = self.sampleContainer["i"][50:57]
        expected.attributes={}
        result.attributes={}
        self.assertEqual(result[0], expected)
        expected = self.sampleContainer["t"][50:57]
        expected.attributes={}
        result.attributes={}
        self.assertEqual(result[1], expected)

    def testANDExpression(self):
        result = self.sampleContainer.filter(
            'col("i") >= "20m" and col("t") <= "98.5s"')
        expectedi = self.sampleContainer["i"][20:98]
        expectedt = self.sampleContainer["t"][20:98]
        expectedi.attributes = {}
        expectedt.attributes = {}
        result[0].attributes = {}
        result[1].attributes = {}
        self.assertEqual(result[0], expectedi)
        self.assertEqual(result[1], expectedt)

    #tests involving 2 dimensional FieldContainers:
    def _compareExpected(self, expression, ind):
        indices = numpy.array(ind)
        result = self.sc2d.filter(expression)
        expectedSC = copy.deepcopy(self.sc2d)
        for FC in expectedSC:
            FC.data = FC.data[indices]
            FC.error = FC.error[indices]
            FC.dimensions[0].data = FC.dimensions[0].data[indices]
        self.assertEqual(result, expectedSC)
        return result

    def testEmpty2dExpression(self):
        result = self.sc2d.filter('')
        self.assertEqual(result, self.sc2d)

    def testAtomar2dExpressions(self):
        self._compareExpected('col("t") <= "40.0s"',
                              [True, True, False, True, False])
        self._compareExpected('col("l") < "10000m"',
                              [True, True, False, False, True])
        self._compareExpected('col("Zeit") >= "20.0s"',
                              [True, True, True, False, True])
        self._compareExpected('col("l") > "5500m"',
                              [False, False, True, True, False])
        self._compareExpected('col("t") == "18000s"',
                              [False, False, False, False, True])
        self._compareExpected('col("Strecke") != "20000m"',
                              [True, True, False, True, True])

    def testNot2dExpression(self):
        self._compareExpected('not col("t") == "10s"',
                              [True, True, True, False, True])

    def testAnd2dExpression(self):
        self._compareExpected(
            'col("Zeit") == "60s" and "20000m" == col("Strecke")',
            [False, False, True, False, False])

    def testMultiAnd2dExpression(self):
        self._compareExpected(
            "col('l') >= '0 km' and col('l') < '20 km' and col('t') <= '60s'",
            [False, True, False, True, False])

    def testOr2dExpression(self):
        self._compareExpected(
            'col("Zeit") < "60s" or col("Strecke") == "5500m"',
            [True, True, False, True, True])

    def testMultiOr2dExpression(self):
        self._compareExpected(
            "col('t') == '20s' or col('l') == '0 km' or col('t') > '1000.2 s'",
            [True, True, False, False, True]
            )

    def testMultipleCompareOpPrecedence2dExpression(self):
        self._compareExpected('not "0m" <= col("l") <= "10000m"',
                              [True, False, True, False, False])

    def testColumnToColumn2dExpression(self):
        self._compareExpected('col("l") == col("Strecke")',
                              [True, True, True, True, True])
        self._compareExpected('col("t") != col("Zeit")',
                              [False, False, False, False, False])

    def testIncompatibleDimensionsExpression(self):
        fc1 = FieldContainer(numpy.array([1, 2, 3]),
                             dimensions=[FieldContainer(numpy.array([4, 5, 6]),
                                                        unit=Quantity('1m'))],
                             shortname='t1')
        fc2 = copy.deepcopy(fc1)
        fc2.dimensions[0].data[0] = 4.01
        fc2.shortname = 't2'
        incompatibleSC = SampleContainer(columns=[fc1, fc2])
        self.assertRaises(ValueError, incompatibleSC.filter,
                          "col('t1') > col('t2')")


class FieldContainerRescaling(unittest.TestCase):
    def setUp(self):
        self.testData = scipy.array([[0,1,2],[3,4,5],[6,7,8]])
        self.longname = u"Sampled Data"
        self.shortname = u"I\\omega"
        self.unit = Quantity('3.14 m')

    def testRescaleUnitless(self):
        field = FieldContainer(copy.deepcopy(self.testData),
                               1,
                               longname=self.longname,
                               shortname=self.shortname)
        nt.assert_array_equal(self.testData,
                              field.data,
                              'Unauthorized rescaling.')
        field = FieldContainer(copy.deepcopy(self.testData),
                               1,
                               longname=self.longname,
                               shortname=self.shortname,
                               rescale=True)
        nt.assert_array_equal(self.testData,
                              field.data,
                              "Rescale option shouldn't do anything for "
                              "unitless fields.")

    def testRescaleBaseUnitsFloats(self):
        field = FieldContainer(copy.deepcopy(self.testData).astype('f'),
                               '1 cm**2/mm',
                               longname=self.longname,
                               shortname=self.shortname,
                               rescale=True)
        nt.assert_array_equal(self.testData,field.data,
                              'Wrong scaling of field.')
        self.assertEqual(field.unit.unit.name(),
                         'dm',
                         'Wrong baseunit: %s.' % field.unit.unit.name())

    def testRescaleFloats(self):
        field = FieldContainer(100 * copy.deepcopy(self.testData).astype('f'),
                               '10 muA',
                               longname=self.longname,
                               shortname=self.shortname,
                               rescale=True)
        nt.assert_array_almost_equal(self.testData,
                                     field.data,5,
                                     'Wrong scaling of field.')
        self.assertEqual(field.unit.unit.name(),
                         'mA',
                         'Wrong baseunit: %s.' % field.unit.unit.name())

    def testRescaleFloatsMeter(self):
        field = FieldContainer(100 * copy.deepcopy(self.testData).astype('f'),
                               '1 cm**2/mm',
                               longname=self.longname,
                               shortname=self.shortname,
                               rescale=True)
        nt.assert_array_almost_equal(10 * self.testData,
                                     field.data,
                                     5,
                                     'Wrong scaling of field.')
        self.assertEqual(field.unit.unit.name(),
                         'm',
                         'Wrong baseunit: %s.' % field.unit.unit.name())

    def testRescaleBaseUnitsInteger(self):
        field = FieldContainer(copy.deepcopy(self.testData),
                               '1 cm**2/mm',
                               longname=self.longname,
                               shortname=self.shortname,
                               rescale=True)
        nt.assert_array_equal(self.testData,
                              field.data,
                              'Wrong scaling of field.')
        self.assertEqual(field.unit.unit.name(),
                         'dm',
                         'Wrong baseunit: %s.' % field.unit.unit.name())

    def testUnitInteger(self):
        field = FieldContainer(copy.deepcopy(self.testData),
                               unit=self.unit,
                               longname=self.longname,
                               shortname=self.shortname,
                               rescale=True)
        self.assertEqual(field.unit,
                         self.unit,
                         "An integer field should not be rescaled, but should "
                         "hold the normation constant in its unit.")

    def testDimensionsRescaling(self):
        """
        Option rescale=True should also rescale the dimensions of the
        FieldContainer.
        """
        dim = [FieldContainer(1000 * scipy.linspace(0, 1, 3),
                              unit='1 nm',
                              longname="Axis %s" % axis,
                              shortname=axis) for axis in ['x','y']]
        field = FieldContainer(copy.deepcopy(self.testData),
                               dimensions=dim,
                               unit=self.unit,
                               longname=self.longname,
                               shortname=self.shortname,
                               rescale=True)
        for dim in [0,1]:
            axisUnit = field.dimensions[dim].unit
            self.assertEqual(axisUnit.unit.name(), 'mum')

class FieldContainerSlicing1d(unittest.TestCase):
    def setUp(self):
        self.field1d = FieldContainer(numpy.linspace(0.1, 1, 10),
                                      longname="voltage",
                                      shortname="U",
                                      unit="1V")
    def testSingleIndex(self):
        section = self.field1d[0]
        afoot = FieldContainer(numpy.array([0.1]),
                               longname="afoot",
                               shortname="U",
                               unit="1V",
                               dimensions=[],
                               attributes={u'Index':(u'i', 0)})
        assertEqual(section, afoot)

    def testRegionIndex(self):
        section = self.field1d[1:4]
        afoot = FieldContainer(numpy.linspace(0.2, 0.4, 3),
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0].data = numpy.linspace(1, 3, 3)
        self.assertEqual(section, afoot)
        section = self.field1d[1:-1]
        afoot = FieldContainer(numpy.linspace(0.2, 0.9, 8),
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0].data = numpy.linspace(1, 8, 8)
        self.assertEqual(section, afoot)

    def testCommaSeparated(self):
        section = self.field1d[[1, 3, 7], ]
        afoot = FieldContainer(numpy.array([0.2, 0.4, 0.8]),
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0].data = numpy.array([1, 3, 7])
        self.assertEqual(section, afoot)
        section = self.field1d[[[1, 3, 7]]]
        self.assertEqual(section, afoot)

    def testRegionIndexUnit(self):
        f = self.field1d.dimensions[0]
        mask = slice(3, 7)
        afoot = FieldContainer(self.field1d.data[mask],
                               dimensions=[f[mask]],
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        self.assertRaises(NotImplementedError,
                          self.field1d.__getitem__,
                          "0.4:0.6")
        section = self.field1d["3:7"]
        assertEqual(section, afoot)


class FieldContainerSlicing1dDim(FieldContainerSlicing1d):
    def setUp(self):
        super(FieldContainerSlicing1dDim, self).setUp()
        self.xDim = FieldContainer(numpy.linspace(0.3,
                                                  0.7,
                                                  self.field1d.data.shape[0]),
                                   longname="current",
                                   shortname="I",
                                   unit="1A")
        self.field1d.dimensions[0] = self.xDim

    def testSingleIndex(self):
        section = self.field1d[0]
        afoot = FieldContainer(numpy.array([0.1]),
                               longname="afoot",
                               shortname="U",
                               unit="1V",
                               dimensions=[self.xDim[0]],
                               attributes={u'current':\
                                           (u'I', Quantity("0.3A"))})
        assertEqual(section, afoot)

    def testRegionIndex(self):
        section = self.field1d[1:4]
        afoot = FieldContainer(numpy.linspace(0.2, 0.4, 3),
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0] = self.xDim[1:4]
        self.assertEqual(section, afoot)
        section = self.field1d[1:-1]
        afoot = FieldContainer(numpy.linspace(0.2, 0.9, 8),
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0] = self.xDim[1:-1]
        self.assertEqual(section, afoot)

    def testCompleteRightOpenIntervall(self):
        dim = self.xDim
        intStart = dim.data.min() * dim.unit
        intEnd = dim.data.max() * dim.unit
        unitname = dim.unit.unit.name()
        section = self.field1d["%.4f%s:%.4f%s" % (intStart.value,
                                                  unitname,
                                                  intEnd.value,
                                                  unitname)]
        afoot = FieldContainer(numpy.linspace(0.1, 0.9, 9),
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0] = self.xDim[0:-1]
        self.assertEqual(section, afoot)

    def testCompleteIntervall(self):
        dim = self.xDim
        intStart = dim.data.min() * dim.unit
        intEnd = dim.data.max() * dim.unit * 1.001
        unitname = dim.unit.unit.name()
        argument = "%.4f%s:%.4f%s" % (intStart.value,
                                      unitname,
                                      intEnd.value,
                                      unitname)
        section = self.field1d[argument]
        self.assertEqual(section, self.field1d)

    def testCommaSeparated(self):
        section = self.field1d[[1, 3, 7], ]
        afoot = FieldContainer(numpy.array([0.2, 0.4, 0.8]),
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0] = self.xDim[[1, 3, 7], ]
        self.assertEqual(section, afoot)
        section = self.field1d[[[1, 3, 7]]]
        self.assertEqual(section, afoot)

    def testRegionIndexUnit(self):
        f = self.field1d.dimensions[0]
        mask = slice(3, 7)
        afoot = FieldContainer(self.field1d.data[mask],
                               dimensions=[f[mask]],
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        section = self.field1d["0.4:0.6A"]
        assertEqual(section, afoot)
        section = self.field1d["400mA:0.6A"]
        assertEqual(section, afoot)

    def testRegionIndexUnitExtendedLeftBoundary(self):
        f = self.field1d.dimensions[0]
        mask = slice(0, 7)
        afoot = FieldContainer(self.field1d.data[mask],
                               dimensions=[f[mask]],
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        section = self.field1d["0.:0.6A"]
        assertEqual(section, afoot)
        section = self.field1d["0mA:0.6A"]
        assertEqual(section, afoot)
        section = self.field1d["300mA:0.6A"]
        assertEqual(section, afoot)

    def testRegionIndexUnitRightBoundary(self):
        f = self.field1d.dimensions[0]
        mask = slice(3, None)
        afoot = FieldContainer(self.field1d.data[mask],
                               dimensions=[f[mask]],
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        section = self.field1d["0.4:1A"]
        assertEqual(section, afoot)
        section = self.field1d["0.4A:1000mA"]
        assertEqual(section, afoot)
        section = self.field1d["0.4A:700mA"]
        assertEqual(section, afoot)


class FieldContainerSlicing2d(unittest.TestCase):
    def setUp(self):
        l = numpy.linspace(0, 0.9, 10)
        m = numpy.meshgrid(l, l * 10)
        self.field2d = FieldContainer(m[0] + m[1],
                                      longname="voltage",
                                      shortname="U",
                                      unit="1V")

    def testSingleIndex(self):
        section = self.field2d[0]
        afoot = FieldContainer(numpy.linspace(0, 0.9, 10),
                               longname="voltage",
                               shortname="U",
                               unit="1V",
                               attributes={u'Index':(u'j', 0)})
        self.assertTrue(section.isValid())
        self.assertEqual(section, afoot)

    def testRegionIndex(self):
        section = self.field2d[1:4]
        afoot = FieldContainer(self.field2d.data[1:4],
                               longname="afoot",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0].data = numpy.linspace(1, 3, 3)
        self.assertEqual(section, afoot)
        section = self.field2d[1:-1]
        afoot = FieldContainer(self.field2d.data[1:-1],
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0].data = numpy.linspace(1, 8, 8)
        self.assertEqual(section, afoot)

    def testCommaSeparated(self):
        section = self.field2d[[1, 3, 7], ]
        afoot = FieldContainer(self.field2d.data[[1, 3, 7], ],
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0].data = numpy.array([1, 3, 7])
        self.assertEqual(section, afoot)
        section = self.field2d[[[1, 3, 7]]]
        self.assertEqual(section, afoot)


class FieldContainerSlicing2dDim(FieldContainerSlicing2d):
    def setUp(self):
        super(FieldContainerSlicing2dDim, self).setUp()
        self.xDim = FieldContainer(numpy.linspace(0.3,
                                                  0.7,
                                                  self.field2d.data.shape[0]),
                                   longname="current",
                                   shortname="I",
                                   unit="1A")
        self.yDim = FieldContainer(numpy.linspace(30,
                                                  70,
                                                  self.field2d.data.shape[0]),
                                   longname="position",
                                   shortname="p",
                                   unit="1cm")
        self.field2d.dimensions = [self.yDim, self.xDim]

    def testSingleIndex(self):
        section = self.field2d[0]
        afoot = FieldContainer(numpy.linspace(0, 0.9, 10),
                               longname="voltage",
                               shortname="U",
                               unit="1V",
                               attributes={u'position':\
                                               (u'p',Quantity("30cm"))},
                               dimensions=[self.xDim])
        self.assertTrue(section.isValid())
        self.assertEqual(section, afoot)

    def testRegionIndex(self):
        section = self.field2d[1:4]
        afoot = FieldContainer(self.field2d.data[1:4],
                               longname="afoot",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0] = self.yDim[1:4]
        afoot.dimensions[1] = self.xDim
        self.assertEqual(section, afoot)
        section = self.field2d[1:-1]
        afoot = FieldContainer(self.field2d.data[1:-1],
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0] = self.yDim[1:-1]
        afoot.dimensions[1] = self.xDim
        self.assertEqual(section, afoot)

    def testCommaSeparated(self):
        section = self.field2d[[1, 3, 7], ]
        afoot = FieldContainer(self.field2d.data[[1, 3, 7], ],
                               longname="voltage",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0] = self.yDim[[1, 3, 7], ]
        afoot.dimensions[1] = self.xDim
        self.assertEqual(section, afoot)
        section = self.field2d[[[1, 3, 7]]]
        self.assertEqual(section, afoot)

    def testRegionIndexUnit(self):
        section = self.field2d["4:6dm"]
        afoot = FieldContainer(self.field2d.data[3:7],
                               longname="afoot",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0] = self.yDim[3:7]
        afoot.dimensions[1] = self.xDim
        self.assertEqual(section, afoot)

    def testRegionIndexWrongUnit(self):
        self.assertRaises(TypeError, self.field2d.__getitem__, "4:6A")

    def test2FoldRegionIndexUnit(self):
        section = self.field2d["4:6dm", "4e-1:.6A"]
        afoot = FieldContainer(self.field2d.data[3:7, 3:7],
                               longname="afoot",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0] = self.yDim[3:7]
        afoot.dimensions[1] = self.xDim[3:7]
        self.assertEqual(section, afoot)

    def test2ndAxisRegionIndexUnit(self):
        section = self.field2d[:, "4e-1:.6A"]
        afoot = FieldContainer(self.field2d.data[:, 3:7],
                               longname="afoot",
                               shortname="U",
                               unit="1V")
        afoot.dimensions[0] = self.yDim
        afoot.dimensions[1] = self.xDim[3:7]
        self.assertEqual(section, afoot)


if __name__ == "__main__":
    #suite = unittest.TestLoader().loadTestsFromTestCase(IsValidFieldContainer)
    #unittest.TextTestRunner().run(suite)
    #suite.addTest(unittest.TestLoader().loadTestsFromTestCase(FieldContainerSlicing1dDim))
    import sys
    if len(sys.argv) == 1:
        unittest.main()
    else:
        suite = unittest.TestLoader().loadTestsFromTestCase(
            eval(sys.argv[1:][0]))
        unittest.TextTestRunner().run(suite)
